from abc import ABC, abstractmethod
import pandas as pd
import json
from arise_predictions.utils import constants, utils
from arise_predictions.preprocessing import job_parser


class JsonJobParser(ABC):

    def __init__(self, columns, start_time_field_name, end_time_field_name):
        self.columns = columns
        self.start_time_field_name = start_time_field_name
        self.end_time_field_name = end_time_field_name

    @abstractmethod
    def get_history_json(self, json_file):
        pass


class DefaultJsonJobParser(JsonJobParser):

    def get_history_json(self, json_file):
        with open(json_file) as json_data:
            data = dict(json.load(json_data))

        execution_data = pd.DataFrame(columns=self.columns)
        for job in data.items():
            row = []
            complete_row = True
            for field in self.columns:
                if field == constants.JOB_DURATION_FIELD_NAME:
                    try:
                        duration = job_parser.get_job_duration_from_json(job[1], self.start_time_field_name,
                                                                         self.end_time_field_name)
                        row.append(duration)
                    except Exception:
                        complete_row = False
                        break  # continue to the jobs loop
                else:
                    row.append(utils.find_item(job[1], field))
            if complete_row:
                row_df = pd.DataFrame([row], columns=self.columns)
                row_df = row_df.dropna(how='all')  # remove 'all nan' rows
                if not row_df.empty:
                    if execution_data.empty:
                        execution_data = row_df
                    else:
                        execution_data = pd.concat([execution_data, row_df], ignore_index=True)

        return execution_data


def find_hierarchy_item(obj, key1, key2):

    metrics_list = obj['metrics']

    for metric in metrics_list:
        if metric["name"] == key1:
            if key2 in metric:
                return metric[key2]
            else:
                return None
    return None


class SAJsonJobParser(JsonJobParser):

    """
    Contains one execution per json file
    """

    def get_history_json(self, json_file):
        with open(json_file) as json_data:
            data = dict(json.load(json_data))

        row = get_history_from_dict(self.columns, data, self.start_time_field_name, self.end_time_field_name)

        if len(row):
            row_df = pd.DataFrame([row], columns=self.columns)
            row_df = row_df.dropna(how='all')  # remove 'all nan' rows
            if not row_df.empty:
                return row_df

        return pd.DataFrame(columns=self.columns)


def get_history_from_dict(columns, data, start_time_field_name=None, end_time_field_name=None):
    row = []
    complete_row = True

    for field in columns:
        if field == constants.JOB_DURATION_FIELD_NAME and start_time_field_name is not None:
            try:
                duration = job_parser.get_job_duration_from_json(data, start_time_field_name, end_time_field_name)
                row.append(duration)
            except Exception:
                complete_row = False
                break  # continue to the jobs loop
        else:
            if '.' in field:
                keys = field.split('.')
                row.append(find_hierarchy_item(data, keys[0], keys[1]))
            else:
                row.append(utils.find_item(data, field))

    return row if complete_row else []


class JsonInCSVJobParser:

    """ Assumes start and end time are given as explicit CSV fields, and all other fields are inside json fields"""

    def __init__(self, columns, start_time_field_name, end_time_field_name):
        self.columns = columns
        self.start_time_field_name = start_time_field_name
        self.end_time_field_name = end_time_field_name

    def get_history_csv(self, raw_df_file_name):

        raw_df = pd.read_csv(raw_df_file_name, sep='\t')

        history_df = pd.DataFrame(columns=self.columns)

        dict_fields = []
        standalone_fields = []

        for col in raw_df.columns.values.tolist():
            if col in self.columns:
                standalone_fields.append(col)
            else:
                try:
                    json.loads(raw_df[col].iloc[0])
                    dict_fields.append(col)
                except Exception:
                    continue

        for index, orig_row in raw_df.iterrows():

            start_time = orig_row[self.start_time_field_name]
            end_time = orig_row[self.end_time_field_name]

            merged_dict = dict()

            for field in dict_fields:
                merged_dict[field] = json.loads(orig_row[field])

            history_row = []

            for field in standalone_fields:
                history_row.append(orig_row[field])

            # remove last column - duration, to be added next from non-json fields
            dict_cols = [field for field in self.columns[:-1] if field not in standalone_fields]
            history_row.extend(get_history_from_dict(dict_cols, merged_dict))

            try:
                duration = utils.get_duration(start_time, end_time)
                if duration == 0:
                    continue # to next orig_row
                else:
                    history_row.append(duration)
            except Exception:
                continue # to next orig_row

            if len(history_row):
                row_df = pd.DataFrame([history_row], columns=self.columns)
                row_df = row_df.dropna(how='all')  # remove 'all nan' rows

                if not row_df.empty:
                    if history_df.empty:
                        history_df = row_df
                    else:
                        history_df = pd.concat([history_df, row_df], ignore_index=True)

        return history_df

class FinetuningMetadataParser:

    def get_metadata_df(self, raw_metadata_file_name):

        raw_df = pd.read_csv(raw_metadata_file_name, sep=',')

        metadata_df = pd.DataFrame()

        dict_fields = []
        standalone_fields = []

        for col in raw_df.columns.values.tolist():
            try:
                json.loads(raw_df[col].iloc[0])
                dict_fields.append(col)
            except Exception:
                standalone_fields.append(col)

        for index, orig_row in raw_df.iterrows():

            metadata_row = []

            for field in standalone_fields:
                metadata_row.append(orig_row[field])

            metadata_dict = get_metadata_from_dict(dict_fields, orig_row)

            metadata_row.extend(metadata_dict.values())

            if len(metadata_row):
                row_df = pd.DataFrame([metadata_row], columns=standalone_fields+list(metadata_dict.keys()))

                if not row_df.empty:
                    if metadata_df.empty:
                        metadata_df = row_df
                    else:
                        metadata_df = pd.concat([metadata_df, row_df], ignore_index=True)

        return metadata_df

def get_metadata_from_dict(dict_fields, row):

    metadata_dict_row = dict()

    for field in dict_fields:
        values = json.loads(row[field])
        for value in values:
            if '==' in value:
                key_value = value.split("==")
                metadata_dict_row[key_value[0]+"_"+field] = key_value[1]

    return metadata_dict_row


import time
from datetime import datetime
import pandas as pd

from arise_predictions.utils import constants
import os
import joblib
import errno
from typing import Any
import logging
import shutil, zipfile

logger = logging.getLogger(__name__)


def find_item(obj, key):
    if key in obj:
        return obj[key]
    for k, v in obj.items():
        if isinstance(v, dict):
            item = find_item(v, key)
            if item:
                return item
    return None


def is_time_format(input_time, time_format):
    try:
        time.strptime(input_time, time_format)
        return True
    except ValueError:
        return False


def from_iso_format(input_time):
    try:
        return datetime.fromisoformat(input_time)
    except ValueError:
        return None


def to_datetime(input_time):
    time_format = get_time_format(input_time)

    if time_format:
        return pd.to_datetime(input_time, format=time_format)
    else:
        return from_iso_format(input_time)


def get_time_format(input_time):
    formats = ['%Y-%m-%d %H:%M:%S', '%m/%d/%Y %I:%M:%S %p EST', '%Y-%m-%d %H:%M:%S+00:00']

    for current_format in formats:
        if is_time_format(input_time, current_format):
            return current_format

    return None


def get_duration(start_time, end_time):
    start_time_formatted = to_datetime(start_time)
    if start_time_formatted:
        end_time_formatted = to_datetime(end_time)
        return pd.Timedelta(end_time_formatted - start_time_formatted).seconds
    else:
        raise Exception("Unsupported time format {}".format(start_time))


def adjust_columns_with_duration(columns, start_time_field_name, end_time_field_name):
    # Replace start and end time columns with duration column

    if not start_time_field_name in columns or not end_time_field_name in columns:
        return columns

    adjusted_columns = columns
    if start_time_field_name in columns:
        adjusted_columns.remove(start_time_field_name)
    if end_time_field_name in columns:
        adjusted_columns.remove(end_time_field_name)
    adjusted_columns.append(constants.JOB_DURATION_FIELD_NAME)
    return adjusted_columns


def get_categorical_features(data):
    # numeric_data = data.select_dtypes(include=np.number) # this doesn't work because some numeric
    # columns are assigned with object types, so we'll have to convert to numeric

    # categorical_features = []

    categorical_features = [cols for cols in data.columns if data[cols].dtype == 'object']

    # for column in data.columns.values.tolist():
    #     if pd.isna(pd.to_numeric(data[column], errors='coerce')).all():
    #         categorical_features.append(column)

    return categorical_features

# TODO add exception handling to all IO functions (OSError, FileNotFoundError)


def write_df_to_csv(df: pd.DataFrame, output_path: str, 
                    output_file: str, index: bool = True) -> str:
    mkdirs(output_path)
    path = os.path.join(output_path, output_file)
    df.to_csv(path, index=index)
    return path


def mkdirs(path: str) -> None:
    """
    Create the specified path, if it does not exist yet.
    """
    if not os.path.exists(path):
        os.makedirs(path)


def persist_estimator(estimator: Any, output_path: str, 
                      output_file: str) -> str:
    """
    Write the estimator (e.g., scikit-learn pipeline) to a pickle file.
    """
    # TODO investigate how to make serialization future-proof
    mkdirs(output_path)
    path = os.path.join(output_path, output_file)
    joblib.dump(estimator, path)
    return path


def get_estimator_file_name(estimator_type: str, estimator_name: str, target_variable: str) -> str:
    return f"estimator-{estimator_type}-{estimator_name}"f"-{target_variable}.pkl"


def load_estimator(input_path: str, pickle_file: str) -> Any:
    """
    Load estimator (e.g., scikit-learn pipeline) from a pickle file.
    """
    path = os.path.join(input_path, pickle_file)

    if not os.path.exists(path):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)
    
    return joblib.load(path)


def add_feature_engineering(metadata_path: str, raw_data: pd.DataFrame, feature_engineering: dict[str, str],
                            metadata_parser_class_name: str) -> pd.DataFrame:
    data = raw_data
    if metadata_parser_class_name:
        logger.info("Parsing jobs metadata with class {}".format(metadata_parser_class_name))
        estimation_module = __import__('preprocessing.custom_job_parser', fromlist=[metadata_parser_class_name])
        class_inst = getattr(estimation_module, metadata_parser_class_name)()
    else:
        class_inst = None
    for fe_name, fe_file in feature_engineering.items():
        if fe_name not in raw_data:
            logger.warning(f'Could not engineer feature {fe_name}. Not such feature.')
            continue
        fe_path = os.path.join(metadata_path, constants.JOB_METADATA_DIR, fe_file)
        if class_inst:
            fe_data = getattr(class_inst, "get_metadata_df")(fe_path)
        else:
            fe_data = pd.read_csv(fe_path, sep=',')
        if fe_data is None or fe_data.empty:
            logger.warning(f'Could not engineer feature {fe_name}. Failed loading {fe_file}.')
            continue
        if fe_name not in fe_data:
            logger.warning(f'Could not engineer feature {fe_name}. Not in {fe_file}.')
            continue
        # add suffix to engineered column names
        suffix = '_' + fe_name
        new_cols = [col if col == fe_name else col + suffix for col in fe_data]
        fe_data.columns = new_cols

        logger.info(f'Merging feature engineering for {fe_name} using:\n{fe_data}')
        new_data = data.merge(fe_data, how='left', on=fe_name, suffixes=(None, suffix))  # extra suffix on overlaps
        new_cols = [col for col in new_data if col not in data]
        numeric_cols = set(new_data.select_dtypes(include='number').columns)
        default_values = {col: 0 if col in numeric_cols else "Other" for col in new_cols}
        # @TODO fill value of numeric col -- mean? -1?
        data = new_data.fillna(default_values)

        logger.info(f'** Done feature engineering for {fe_name} using:\n{fe_data}')
    return data

def get_unpacked_path(input_path):
    if zipfile.is_zipfile(input_path):
        input_folder = os.path.join(os.path.dirname(input_path), os.path.splitext(os.path.basename(
            input_path))[0])
        shutil.unpack_archive(input_path, input_folder)
        return input_folder
    else:
        return input_path


import unittest
import logging
import sys
import os

import pandas as pd

from tests.utils.logger_redirector import LoggerRedirector

"""
Tests to determine which method to use for identifying categorical variables.
"""

logger = logging.getLogger('test-logger')

logging.basicConfig(
    format="%(asctime)s %(module)s %(levelname)s: %(message)s",
    level=logging.INFO,
    stream=sys.stdout)


class TestCatId(unittest.TestCase):

    def setUp(self):
        # unittest has reassigned sys.stdout and sys.stderr by this point
        LoggerRedirector.redirect_loggers(fake_stdout=sys.stdout, 
                                          fake_stderr=sys.stderr) 
        
        self.resources_path = "tests/resources"
        self.data_file = "mlcommons_inference_data_execution_history.csv"
        
        self.cat_vars = ['Accelerator', 'Availability', 'Model MLC', 'Organization', 'Processor', 'Scenario',
                         'System Name']

    def tearDown(self):
        LoggerRedirector.reset_loggers(
            fake_stdout=sys.stdout, fake_stderr=sys.stderr)
    
        # if os.path.exists(self.output_path):
        #     shutil.rmtree(self.output_path)
    
    def test_get_categorical_by_object_dtype(self):
        """
        Test that we identify all categorical features by 'object' dtype.
        """
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        categorical_by_obj = [
            cols for cols in data.columns if data[cols].dtype == 'object']
        logger.info(("categorical features by 'object' dtype:" 
                     f" {categorical_by_obj}"))
        self.assertListEqual(categorical_by_obj, self.cat_vars)

    def test_get_categorical_by_numeric_data_function(self):
        """
        Test we identify all categorival features using _get_numeric_data().
        Probably not a good idea to use a hidden function, but one of the
        recommendations on the Interwebs
        """
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        cols = data.columns
        num_cols = data._get_numeric_data().columns
        categorical_by_elimination = list(set(cols) - set(num_cols))
        logger.info(("categorical features by elimination:" 
                     f" {categorical_by_elimination}"))
        categorical_by_elimination.sort()
        self.assertListEqual(categorical_by_elimination, self.cat_vars)

    def test_get_categorical_by_failed_conversion_to_numeric(self):
        """
        Test that we identify all categorical features by failed ('NaN') 
        numeric conversion.

        Note this test fails as the method fails to detect one of the
        categorical variables as such (quantization mode). Can decide to negate
        the test condition to make this pass in case we run automated test
        suites at some point.
        """
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        categorical_by_num = []

        for column in data.columns.values.tolist():
            if pd.isna(pd.to_numeric(data[column], errors='coerce')).all():
                categorical_by_num.append(column)
        logger.info(("categorical features by conversion to numeric:" 
                     f" {categorical_by_num}"))
        self.assertListEqual(categorical_by_num, self.cat_vars)


import argparse

from typing import Sequence

"""
This module parses command line arguments and passes the information to the main execution program
"""

cmd_args = "test123"


def parse_args(argv: Sequence[str] = None):
    """
    Parses the command line arguments and returns the parsed arguments.

    :param argv: Optional sequence of command-line arguments. Allows unit tests
        to pretend that a certain set of command-line arguments have been
        supplied. Will be ignored if not set.
    :type argv: Sequence
    """
    # Create the argument parser object
    parser = argparse.ArgumentParser(
        description='Command-line argument parser')
    parser.add_argument('--loglevel', '-l', default='debug',
                        help='log level', nargs='?', type=str)
    parser.add_argument('--num-jobs', default=-1,
                        help='number of cores to use. default is all', nargs='?', type=int)
    subparsers = parser.add_subparsers(
        help='commands help', dest='command')

    # parser for the preprocess command
    parser_preprocess = subparsers.add_parser(
        'preprocess', help='preprocess help')
    parser_preprocess.add_argument(
        '--input-path', default='examples/MLCommons', help='input path', nargs='?', type=str)
    parser_preprocess.add_argument(
        '--reread-history', help='reread history', action='store_true')
    parser_preprocess.add_argument(
        '--job-file', help='new job file full path for relevant history', nargs='?', type=str)
    parser_preprocess.add_argument(
        '--input-file', help='input data file name, in case only one of the files in input path should be used',
        nargs='?', type=str)
    parser_preprocess.add_argument(
        '--ignore-metadata', help='ignore metadata if exists', action='store_true')

    # parser for analyze jobs command
    parser_analyze_jobs = subparsers.add_parser(
        'analyze-jobs', help='analyze jobs help')
    parser_analyze_jobs.add_argument(
        '--input-path', default='examples/MLCommons', help='input path', nargs='?', type=str)
    parser_analyze_jobs.add_argument(
        '--reread-history', help='reread history', action='store_true')
    parser_analyze_jobs.add_argument(
        '--job-file', help='new job file full path for relevant history', nargs='?', type=str)
    parser_analyze_jobs.add_argument(
        '--input-file', help='input data file name, in case only one of the files in input path should be used',
        nargs='?', type=str)
    parser_analyze_jobs.add_argument(
        '--job-id-column', help='column name representing job id column. If not given, custom job name will be used '
                                'instead.',
        required=False, type=str)
    parser_analyze_jobs.add_argument(
        '--custom-job-name', help='job name to be used in job id column in case none is specified in job id column. '
                                  'If not given, default job name will be used.',
        required=False, type=str)
    parser_analyze_jobs.add_argument(
        '--ignore-metadata', help='ignore metadata if exists', action='store_true')

    # parser for auto build models command
    parser_auto_build_models = subparsers.add_parser(
        'auto-build-models', help='auto build models help')
    parser_auto_build_models.add_argument(
        '--input-path', default='examples/MLCommons', help='input path', nargs='?', type=str)
    parser_auto_build_models.add_argument(
        '--reread-history', help='reread history', action='store_true')
    parser_auto_build_models.add_argument(
        '--job-file', help='new job file full path for relevant history', nargs='?', type=str)
    parser_auto_build_models.add_argument(
        '--input-file', help='input data file name, in case only one of the files in input path should be used',
        nargs='?', type=str)
    parser_auto_build_models.add_argument(
        '--config-file', default='config/default-auto-model-search-config.yaml', 
        help="path to config file defining estimators and parameter search space",
        required=False, type=str)
    parser_auto_build_models.add_argument(
        '--leave-one-out-cv', help='leave one group out cross validation. Provide list of feature names separated'
                                   ' by commas', required=False, type=str)
    parser_auto_build_models.add_argument(
        '--randomized-hpo', help='Use random sampling hyperparameter optimization', action='store_true')
    parser_auto_build_models.add_argument(
        '--random-iterations', help='Number of sampling iterations to perform for each model hyper-parameter space',
        required=False, type=int)
    parser_auto_build_models.add_argument(
        '--feature-column', help='Name of the feature to extrapolate on',
        required=False, type=str)
    parser_auto_build_models.add_argument(
        '--low-threshold', help='exclude feature values less than or equal to the threshold',
        required=False, type=int)
    parser_auto_build_models.add_argument(
        '--high-threshold', help='exclude feature values greater than or equal to the threshold',
        required=False, type=int)
    parser_auto_build_models.add_argument(
        '--ignore-metadata', help='ignore metadata if exists', action='store_true')
    parser_auto_build_models.add_argument(
        '--single-output-file', help='package output into a single file', action='store_true')

    # parser for demo-predict command
    parser_demo_predict = subparsers.add_parser(
        'demo-predict', help='demo predict help')
    parser_demo_predict.add_argument(
        '--input-path', default='examples/MLCommons', help='input path', nargs='?', type=str)
    parser_demo_predict.add_argument(
        '--reread-history', help='reread history', action='store_true')
    parser_demo_predict.add_argument(
        '--input-file', help='input data file name, in case only one of the files in input path should be used',
        nargs='?', type=str)
    parser_demo_predict.add_argument(
        '--config-file', default='config/example-demo-mlcommons-config.yaml',
        help="path to config file defining estimators, target variables, and input search space",
        required=False, type=str)
    parser_demo_predict.add_argument(
        '--model-path', 
        help="path to serialized estimators", required=True, type=str)
    parser_demo_predict.add_argument(
        '--ignore-metadata', help='ignore metadata if exists', action='store_true')

    # parser for predict command
    parser_predict = subparsers.add_parser('predict', help='predict help')
    parser_predict.add_argument(
        '--input-path', default='', help='input path', nargs='?', type=str)
    parser_predict.add_argument(
        '--config-file', default='config/example-demo-mlcommons-config.yaml',
        help="path to config file defining estimators, target variables, and input search space",
        required=False, type=str)
    parser_predict.add_argument(
        '--model-path', help="path to serialized estimators", required=True, type=str)
    parser_predict.add_argument(
        '--ignore-metadata', help='ignore metadata if exists', action='store_true')

    # parser for data-predict command
    parser_data_predict = subparsers.add_parser(
        'data-predict', help='data predict help')
    parser_data_predict.add_argument(
        '--original-data-file', help='file containing the original data', nargs='?', required=False, type=str)
    parser_data_predict.add_argument(
        '--prediction-data-file', help='file containing the prediction data', required=True, nargs='?', type=str)
    parser_data_predict.add_argument(
        '--delta-only', help='predict only for prediction data not in original data', action='store_true')
    parser_data_predict.add_argument(
        '--input-path', help='input path', nargs='?', type=str)
    parser_data_predict.add_argument(
        '--config-file', default='config/example-demo-mlcommons-config.yaml',
        help="path to config file defining estimators, target variables, and input search space", type=str)
    parser_data_predict.add_argument(
        '--model-path', help="path to serialized estimators", required=True, type=str)

    # Parse the command line arguments
    global cmd_args
    if argv: 
        cmd_args = parser.parse_args(argv)
    else:
        cmd_args = parser.parse_args()
    return True


def get_args():
    global cmd_args
    return cmd_args



from typing import Dict, Any
from dataclasses import dataclass
import logging
from itertools import product
import os
import sys
import glob
from functools import reduce
import shutil, zipfile
import yaml

import pandas as pd

from sklearn.metrics import mean_absolute_percentage_error

from arise_predictions.utils import utils, constants

logger = logging.getLogger(__name__)

"""
Performs predictions on input data using previously build estimators and
configuration defining space of input variables to explore. Processes historic
data so that it can create comparisons with ground truth as much as possible for
demo purposes.
"""


@dataclass
class PredictionInputSpace:

    fixed_values: list[Dict[str, Any]]
    data_values: list[Dict[Any, Any]]
    variable_values: list[Dict[str, Any]]
    interpolation_values: list[str]
    estimators: list[Dict[str, Any]]


def get_predict_config(config_file: str) -> PredictionInputSpace:
    logger.info(f"Reading YAML configuration: {config_file}")
    try:
        with open(config_file, "r") as f:
            config_dict = yaml.safe_load(f)
    except FileNotFoundError:
        raise ValueError(f"Configuration file not found: {config_file}")
    except yaml.YAMLError as e:
        raise ValueError(f"Error parsing configuration file: {e}")
    return get_predict_config_from_dict(config_dict)


def get_predict_config_from_dict(config_dict: dict[str, list[dict[Any, Any]]]) -> PredictionInputSpace:
    return PredictionInputSpace(config_dict.get(constants.PRED_CONFIG_FIXED, []),
                                config_dict.get(constants.PRED_CONFIG_DATA, []),
                                config_dict.get(constants.PRED_CONFIG_VARIABLE, []),
                                config_dict.get(constants.PRED_CONFIG_INTERPOLATION, []),
                                config_dict[constants.PRED_CONFIG_ESTIMATORS])


def _create_input_space(input_config: PredictionInputSpace,
                        original_data: pd.DataFrame,
                        output_path: str, feature_engineering: dict[str, str] = None,
                        metadata_parser_class_name: str = None, metadata_path: str = None) -> tuple[str, pd.DataFrame]:
    """
    Takes the configuration of the input space and creates a corresponding
    DataFrame to run predictions on.

    :param input_config: Parsed input feature space configuration
    :type input_config: PredictionInputSpace
    :param original_data: Original data used to create the model
    :type original_data: pd.DataFrame
    :param output_path: Path under which to store outputs of this script
    :type output_path: str
    :param feature_engineering: map from feature to its metadata features
    :type feature_engineering: dict[str, str]
    :param metadata_parser_class_name: class name of metadata features parser
    :type metadata_parser_class_name: str
    :param metadata_path: path to metadata files
    :type metadata_path: str
    :returns: Tuple containing path under which DataFrame representing input 
        feature space has been stored and the corresponding data frame
    :rtype: tuple[str, pd.DataFrame]
    """
    logger.info(f"Creating input feature space")
    fixed_values = {k: v for d in input_config.fixed_values for k, v in d.items()}
    data_values = {k[constants.PRED_CONFIG_DATA_INPUT]: _get_data_range_values(original_data, k[constants.PRED_CONFIG_DATA_INPUT], k)
                   for k in input_config.data_values} if original_data is not None and not original_data.empty else {}
    variable_values = {k: v for d in input_config.variable_values for k, v in d.items()}
    interpolation_values = {k: _get_data_range_missing_values(original_data, k) for k in
                            input_config.interpolation_values} if original_data is not None and not \
        original_data.empty else {}

    if variable_values.keys() & fixed_values.keys():
        logger.error("Cannot specify same variable name as variable values and fixed value")
        raise Exception

    if variable_values.keys() & interpolation_values.keys():
        logger.error("Cannot specify same variable name as variable values and interpolation values")
        raise Exception

    non_fixed_values = dict(data_values)
    for k, v in variable_values.items():
        non_fixed_values[k] = non_fixed_values[k] + v if k in data_values.keys() else v
    variable_dict = non_fixed_values | interpolation_values

    # create Cartesian product of variable values (much nicer than nested loops)
    combinations = list(product(*(variable_dict.values())))

    # create DF with resulting combinations
    input_space_df = pd.DataFrame(combinations, columns=variable_dict.keys())

    # add the fixed values to each row in that DF
    for var_name, val in fixed_values.items():
        input_space_df[var_name] = val

    if feature_engineering:
        input_space_df = utils.add_feature_engineering(metadata_path, input_space_df, feature_engineering,
                                                       metadata_parser_class_name)

    output_file_df = constants.PRED_INPUT_SPACE_FILE

    path = utils.write_df_to_csv(
        df=input_space_df,
        output_path=output_path,
        output_file=output_file_df)
    logger.info(f"Wrote input feature space to {path}")
    return path, input_space_df


def _get_data_range_missing_values(original_data: pd.DataFrame, var_name: str) -> list[str]:
    values = range(original_data[var_name].min() + 1, original_data[var_name].max())
    return [val for val in values if val not in original_data[var_name].tolist()]


def _get_data_range_values(original_data: pd.DataFrame, var_name: str, var_info: Dict[Any, Any]) -> list[str]:
    values = original_data[var_name].dropna().unique().tolist() if \
        var_info[constants.PRED_CONFIG_DATA_VALUES] == constants.PRED_CONFIG_DATA_ALL else \
        list(range(original_data[var_name].min(), original_data[var_name].max()+1))
    return values if constants.PRED_CONFIG_DATA_EXCLUDE not in var_info.keys() else \
        [val for val in values if val not in var_info[constants.PRED_CONFIG_DATA_EXCLUDE]]


def _get_highest_ranked_estimator(estimator_path: str, target_variable: str) -> str:
    ranking_file = os.path.join(estimator_path, constants.AM_RANKINGS_FILE)
    if not os.path.exists(ranking_file):
        logger.error("Failed to locate model ranking file")
        raise Exception
    ranking_df = pd.read_csv(ranking_file)
    ranking_df_var = ranking_df.loc[ranking_df[constants.AM_COL_TARGET] == target_variable]
    ranking_df_var['combined_rank'] = ranking_df_var[constants.AM_COL_RANK_MAPE] + \
                                      ranking_df_var[constants.AM_COL_RANK_NRMSE_MAXMIN]
    best_row = ranking_df_var.loc[[ranking_df_var['combined_rank'].idxmin()]]
    estimator_type = "linear" if best_row[constants.AM_COL_LINEAR].values[0] else "nonlinear"
    return utils.get_estimator_file_name(estimator_type, best_row[constants.AM_COL_ESTIMATOR].values[0],
                                         target_variable)


def _run_predictions(original_data: pd.DataFrame,
                     input_data: pd.DataFrame,
                     estimators_config: list[Dict[str, Any]],
                     estimator_path: str,
                     output_path: str) -> None:
    """"
    For each target variable in the estimator configuration, load the
    corresponding estimator and run predictions on the input data. Also merges
    all predictions into a single CSV file to make it easier to evaluate
    trade-offs between target variables.
    """
    logger.info("Running predictions")
    target_variables = []

    estimator_folder = utils.get_unpacked_path(estimator_path)

    # run predictions
    for entry in estimators_config:
        target_variable = entry[constants.PRED_CONFIG_TARGET_VAR]
        target_variables.append(target_variable)
        greater_is_better = entry[constants.PRED_CONFIG_GREATER_BETTER]
        if constants.PRED_CONFIG_ESTIMATOR_FILE in entry:
            estimator_file = entry[constants.PRED_CONFIG_ESTIMATOR_FILE]
        else:
            # We consult ranking for each target variable separately, because for some
            # target variables the user mau specify an estimator, while for others not
            # (in which one the best one according to ranking will be taken)
            estimator_file = _get_highest_ranked_estimator(estimator_folder, target_variable)
        logger.info((f"Predicting target variable {target_variable}"
                     f" with {estimator_file}"))
        estimator = utils.load_estimator(
            input_path=estimator_folder, pickle_file=estimator_file)
        y_pred = estimator.predict(input_data)
        predictions_df = input_data.copy(deep=True)
        predictions_df[target_variable] = y_pred

        predictions_ranked_df = _rank_predictions(
            data=predictions_df,
            target_column=target_variable,
            greater_is_better=greater_is_better
        )

        output_file_preds = f"predictions-{target_variable}.csv"
        path = utils.write_df_to_csv(
            df=predictions_ranked_df,
            output_path=output_path,
            output_file=output_file_preds)
        logger.info(f"Wrote predictions to {path}")

    # merge predictions to facilitate demos
    logger.info("Merging predictions")
    all_predictions_files = glob.glob(os.path.join(
        output_path,
        r"predictions-*.csv"))
    dfs = []

    for filename in all_predictions_files:
        # TODO can find better way to handle/avoid this with more time
        ignore_files = [
            constants.PRED_ALL_PREDICTIONS_FILE,
            constants.PRED_GROUND_TRUTH_FILE]
        if any(x in filename for x in ignore_files):
            continue
        df = pd.read_csv(filename, index_col=None, header=0)
        dfs.append(df)
    all_predictions_df = reduce(lambda x, y: pd.merge(x, y), dfs)
    all_predictions_df = all_predictions_df.loc[:, ~all_predictions_df.columns.str.contains('^Unnamed')]
    output_file_all_predictions = constants.PRED_ALL_PREDICTIONS_FILE
    path = utils.write_df_to_csv(
        df=all_predictions_df,
        output_path=output_path,
        output_file=output_file_all_predictions, index=False)
    logger.info(f"Wrote merged predictions to {path}")

    # create comparison with original data

    if original_data is not None and not original_data.empty:
        merged_df = pd.merge(
            original_data,
            all_predictions_df,
            on=input_data.columns.values.tolist(),
            suffixes=["_actual", "_pred"])
        if not merged_df.empty:
            merged_df = merged_df.loc[:, ~merged_df.columns.str.contains('^Unnamed')]
            merged_params = merged_df.columns.values.tolist()
            for p in all_predictions_df:
                if p + "_pred" in merged_params and p + "_actual" in merged_params:
                    merged_df[p + "_mape"] = [mean_absolute_percentage_error([y_t], [y_p]) for y_t, y_p in
                                              zip(merged_df[p + "_actual"], merged_df[p + "_pred"])]
            output_file_merged = constants.PRED_GROUND_TRUTH_FILE
            path = utils.write_df_to_csv(
                df=merged_df,
                output_path=output_path,
                output_file=output_file_merged, index=False)
            logger.info(f"Wrote source of truth to {path}")

            path = utils.write_df_to_csv(
                df=original_data,
                output_path=output_path,
                output_file=constants.PRED_ORIGINAL_TRUTH_FILE, index=False)
            logger.info(f"Wrote original data to {path}")

    if zipfile.is_zipfile(estimator_path):
        # remove temporary estimator artifacts folder created from given archive file
        shutil.rmtree(estimator_folder)


def _rank_predictions(
        data: pd.DataFrame,
        target_column: str,
        greater_is_better: bool = False) -> pd.DataFrame:
    """
    Rank the indicated target variable column according to whether larger 
    or smaller values are preferred. Modifies DataFrame inplace by adding
    suitable ranking column.

    :param data: DataFrame containing predictions.
    :type data: pandas.DataFrame
    :param target_column: Name of the column whose values should be ranked.
        Assumes numeric data.
    :type target_columns: str
    :param greater_is_better: Indicates whether larger or smaller values are
        considered better.
    :type: bool
    :returns: DataFrame with ranking column for each target column.
    """
    data.insert(
        loc=len(data.columns),
        column=f"rank_{target_column}",
        value=data[target_column].rank(ascending=not greater_is_better))
    return data


def demo_predict(original_data: pd.DataFrame, config: PredictionInputSpace,
                 estimator_path: str, feature_engineering: dict[str, str] = None,
                 metadata_parser_class_name: str = None, metadata_path: str = None, output_path: str = None):
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    if output_path is None:
        msg = "Must set output_path for demo predict outputs"
        logger.error(msg)
        raise ValueError(msg)

    logger.info("Beginning demo predict")
    _, input_space_df = _create_input_space(
        input_config=config,
        original_data=original_data,
        output_path=output_path,
        feature_engineering=feature_engineering,
        metadata_parser_class_name=metadata_parser_class_name,
        metadata_path=metadata_path
    )
    _run_predictions(
        original_data=original_data,
        input_data=input_space_df,
        estimators_config=config.estimators,
        estimator_path=estimator_path,
        output_path=output_path
    )
    logger.info(f"Demo predict outputs written to {output_path}")

# Predict outputs for every row in a given prediction_data dataframe. If original data ia given, accuracy is computed.
# If delta_only is True, predictions are performed only for the delta between the prediction data and the original data.


def data_predict(original_data: pd.DataFrame, prediction_data: pd.DataFrame, estimator_path: str,
                 estimators_config: list[Dict[str, Any]], target_variables: list[str], delta_only: bool = False,
                 output_path: str = None):
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    if output_path is None:
        msg = "Must set output_path for demo predict outputs"
        logger.error(msg)
        raise ValueError(msg)

    prediction_data_inputs = prediction_data[[c for c in prediction_data.columns if c not in target_variables]]
    original_data_inputs = original_data[[c for c in original_data.columns if c not in target_variables]] if \
        original_data is not None and not original_data.empty else None

    logger.info("Beginning data predict")
    _run_predictions(
        original_data=original_data,
        input_data=prediction_data_inputs[~prediction_data.isin(original_data_inputs)].dropna(how='any') if delta_only
        and original_data_inputs is not None and not original_data_inputs.empty else prediction_data_inputs,
        estimators_config=estimators_config,
        estimator_path=estimator_path,
        output_path=output_path
    )
    logger.info(f"Demo predict outputs written to {output_path}")


import unittest
import pandas as pd
import os
import sys
import logging
import shutil

from arise_predictions.job_statistics.analyze_jobs import \
    (_init_args,
     _prepare_data,
     _compute_missing_data,
     _summarize_data,
     _correlate_data,
     _summarize_categorical_data,
     _plot_histograms,
     _plot_box_whiskers,
     _plot_correlations)
from arise_predictions.utils import constants
from tests.utils.logger_redirector import LoggerRedirector

"""
Tests for analyze data functionality.
"""

logger = logging.getLogger('test-logger')

logging.basicConfig(
    format="%(asctime)s %(module)s %(levelname)s: %(message)s",
    level=logging.INFO,
    stream=sys.stdout)


class TestAnalyzeData(unittest.TestCase):

    def setUp(self):
        # unittest has reassigned sys.stdout and sys.stderr by this point
        LoggerRedirector.redirect_loggers(fake_stdout=sys.stdout, 
                                          fake_stderr=sys.stderr)

        self.output_path = "test-tmp/data-analysis"
        self.job_id_column = "Custom_Job_Column"  # from job_spec.yaml

        self.full_row_job_name_1 = ["job-1", 50, 100, "table-1", 101]
        self.full_row_job_name_2 = ["job-1", 60, 90, "table-1", 303]
        self.full_row_job_name_3 = ["job-2", 60, 200, "table-1", 99]
        self.full_row_job_name_4 = ["job-2", 70, None, "table-3", 909]
        self.full_row_job_name_5 = ["job-2", 80, None, "table-3", 900]
        self.full_row_job_name_6 = ["job-2", 90, 300, "table-3", 919]
        self.empty_row_job_name_1 = [None, None, None, None, None]

        self.input_data_with_job_id = pd.DataFrame(
            [self.full_row_job_name_1, self.full_row_job_name_2,
                self.full_row_job_name_3, self.full_row_job_name_4,
                self.full_row_job_name_5, self.full_row_job_name_6,
                self.empty_row_job_name_1],
            columns=[
                self.job_id_column,  "Metric-1", "Metric-2", "Table", 
                "duration"])

        self.missing_values_file = "missing-values-job-2.csv"
        self.summary_file = "descriptive-stats.csv"
        self.correlation_file_csv_job_1 = "correlations-job-1.csv"
        self.correlation_file_csv_job_2 = "correlations-job-2.csv"
        self.histogram_file_job_1 = "histogram-job-1.pdf"
        self.histogram_file_job_2 = "histogram-job-2.pdf"
        self.box_file_job_1 = "box-job-1.pdf"
        self.box_file_job_2 = "box-job-2.pdf"
        self.correlation_file_job_1 = "correlation-job-1.png"
        self.correlation_file_job_2 = "correlation-job-2.png"
        self.categorical_table_metric1_file = "categorical-stats-Metric-1.csv"
        self.categorical_table_metric2_file = "categorical-stats-Metric-2.csv"
        self.categorical_table_duration_file = "categorical-stats-duration.csv"
        
        self.full_row_no_job_name_1 = [500, 100, None, 101]
        self.full_row_no_job_name_2 = [600, 200, None, 303]
        self.empty_row_no_job_name_1 = [None, None, None, None]

        self.input_data_without_job_name = pd.DataFrame(
            [self.full_row_no_job_name_1, self.full_row_no_job_name_2, 
             self.empty_row_no_job_name_1],
            columns=["Metric-1", "Metric-2", "Metric-3", "duration"])

        self.custom_job_name = "Custom_Job_Name"

    def tearDown(self):
        LoggerRedirector.reset_loggers(
            fake_stdout=sys.stdout, fake_stderr=sys.stderr)
        # unittest will revert sys.stdout and sys.stderr after this

        if os.path.exists(self.output_path):
            shutil.rmtree(self.output_path)

    # Data preparation tests
        
    def test_drops_empty_rows_and_columns(self):
        """
        All empty rows and empty columns are dropped.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id, 
            job_id_column=self.job_id_column, 
            custom_job_name=None)
        self.assertEqual(len(result_data.index), 6)
        self.assertNotIn("Aggregate", result_data.columns)
    
    def test_job_name_column_from_input_data_exists(self):
        """
        The job id column from the preprocessed input data exists in the 
        resulting DataFrame.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column, 
            custom_job_name=None)
        self.assertIn(self.job_id_column, result_data.columns)

    def test_job_name_column_is_added(self):
        """
        If no job id column exists in preprocessed data (i.e. not specified in
        job_spec.yaml) and no custom job name is specified via an argument, 
        the initialization of arguments will return the default job name column
        and default job name. prepare_data will add this column and value to 
        the data.
        """
        job_id_column, job_name = _init_args(
            job_id_column=None, 
            custom_job_name=None, 
            output_path=self.output_path)
        result_data = _prepare_data(
            data=self.input_data_without_job_name,
            job_id_column=job_id_column, 
            custom_job_name=job_name)
        self.assertIn(constants.DEFAULT_JOB_ID_COLUMN, result_data.columns)

    def test_job_name_column_with_job_name_is_added(self):
        """"
        If no job id column exists in preprocessed data (i.e. not speicified in
        job_spec.yaml), but a custom job name is specified via an argument, a
        default job name column filled with this job name will be added.
        """
        job_id_column, job_name = _init_args(
            job_id_column=None,
            custom_job_name=self.custom_job_name,
            output_path=self.output_path)
        result_data = _prepare_data(
            data=self.input_data_without_job_name,
            job_id_column=job_id_column,
            custom_job_name=job_name)
        self.assertIn(constants.DEFAULT_JOB_ID_COLUMN, result_data.columns)
        self.assertIn(
            self.custom_job_name, 
            result_data[constants.DEFAULT_JOB_ID_COLUMN].values)
    
    def test_job_name_param_ignored_if_job_col_exists(self):
        """
        A custom a job name parameter is ignored, if the data contains a
        job id column that is known to it (e.g., from job_spec.yaml).
        """
        job_id_column, job_name = _init_args(
            job_id_column=self.job_id_column,
            custom_job_name="TESTJOB",
            output_path=self.output_path)
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=job_id_column,
            custom_job_name=job_name)
        self.assertIn(self.job_id_column, result_data.columns)
        self.assertNotIn("TESTJOB", result_data[self.job_id_column].values)
        
    # Null values tests

    def test_identifies_null_values(self):
        """
        Missing values are identified correctly.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column,
            custom_job_name=None)
        _compute_missing_data(
            data=result_data, 
            job_id_column=self.job_id_column, 
            output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.missing_values_file)))
        result_df = pd.read_csv(os.path.join(
            self.output_path, self.missing_values_file))
        self.assertEqual(result_df.Percentage[0], 50.0)
    
    # Summary statistics tests

    def test_summary_statistics_output_exists(self):
        """
        Summary statistics CSV file is created as expected.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column,
            custom_job_name=None) 
        _summarize_data(
            data=result_data, 
            job_id_column=self.job_id_column, 
            output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.summary_file)))
    
    def test_correlation_analysis_output_exists(self):
        """
        Files for correlation matrix plot and numeric representation is created
        as expected.    
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column,
            custom_job_name=None) 
        _correlate_data(
            data=result_data,
            job_id_column=self.job_id_column,
            output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.correlation_file_csv_job_1)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.correlation_file_csv_job_2)))
    
    def test_categorical_summary_output_exists(self):
        """
        Files for target variable aggregations by categorical variables exist.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column,
            custom_job_name=None) 
        _summarize_categorical_data(
            data=result_data,
            target_variables=["Metric-1", "Metric-2"],
            output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.categorical_table_metric1_file)))
    
    # Visualization tests

    def test_histogram_output_exists(self):
        """
        The image files containing histograms have been created.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column,
            custom_job_name=None) 
        _plot_histograms(
            data=result_data,
            job_id_column=self.job_id_column,
            output_path=self.output_path) 
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.histogram_file_job_1)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.histogram_file_job_2)))
        
    def test_box_whiskers_output_exists(self):
        """
        The images files containing box-whisker plots have been created.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column,
            custom_job_name=None) 
        _plot_box_whiskers(
            data=result_data,
            job_id_column=self.job_id_column,
            output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.box_file_job_1)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.box_file_job_2)))
    
    def test_correlation_matrix_exists(self):
        """
        The image files containing correlation matrices have been created.
        """
        result_data = _prepare_data(
            data=self.input_data_with_job_id,
            job_id_column=self.job_id_column,
            custom_job_name=None) 
        _plot_correlations(
            data=result_data,
            job_id_column=self.job_id_column,
            output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.correlation_file_job_1)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.correlation_file_job_2)))
       

if __name__ == "__main__":
    unittest.main()




import sys
import pandas as pd
import logging
import os
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np

from arise_predictions.utils import constants, utils

logger = logging.getLogger(__name__)

"""
Analyzes task/job measurements and prepares various descriptive statistics in
numeric and diagramatic form.
"""


def _init_args(job_id_column=None, custom_job_name=None,
               output_path=None) -> tuple[str, str]:
    """
    :param job_id_column: Name of the column containing the job name.
    Is referred to as 'job id' in job_spec.yaml. 
    :type job_id_column: str
    :param job_name: Name to use for value in job id column. Allows
    user to specify job name if none originally contained in data or our code
    to add a default one so that code can assume a job_id_column.
    :type job_name: str
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    :returns: Name of job id column and custom job name
    :rtype: tupele[str, str]
    """
    if job_id_column is None:
        job_id_column = constants.DEFAULT_JOB_ID_COLUMN
    job_id_column: str = job_id_column

    if custom_job_name is None:
        custom_job_name = constants.DEFAULT_JOB_NAME
    custom_job_name: str = custom_job_name

    if output_path is None:
        msg = "Must set output_path for data analysis results"
        logger.error(msg)
        raise ValueError(msg)

    # fontsize of the axes title
    plt.rc('axes', titlesize=constants.FONT_SMALL_SIZE)     
    plt.rc('figure', titlesize=constants.FONT_MEDIUM_SIZE)

    return job_id_column, custom_job_name


def _prepare_data(data: pd.DataFrame, job_id_column: str, 
                  custom_job_name: str) -> pd.DataFrame:
    """
    Prepare the preprocessed measurements data for subsequent analysis
    procedures.
    :param data: The preprocessed measurements.
    :type data: pandas.DataFrame
    :param job_id_column: Name of column containing job ids
    :type job_id_column: str
    :param custom_job_name: Optional name of job name to insert in case none
    exists in data execution history data frame. Enables job analysis code to 
    work the same regardless of whether or not job ids are in data.
    :type custom_job_name: str
    :returns: Data execution history ready for data analysis
    :rtype: pandas.DataFrame
    """
    logger.info("Preparing data for analysis")

    # drop rows without data
    data = data.dropna(how='all')

    # drop columns where all rows are without data
    data = data.dropna(axis=1, how='all')

    # add task_name column if does not exist
    if job_id_column not in data.columns:
        data.insert(0, job_id_column, custom_job_name)

    # convert numeric rows to numeric
    cols = [i for i in data.columns if data[i].dtype.kind in "iufc"]
    for col in cols:
        data[col] = pd.to_numeric(data[col])
    return data
    

def _compute_missing_data(data: pd.DataFrame, job_id_column: str,
                          output_path: str) -> None:
    """"
    Computes statistics about missing values per column/attribute and 
    serializes to a CSV file. 

    :param data: Data execution history data preprocessed from
    measurements and prepared by _prepare_data.
    :type data: pandas.DataFrame
    :param job_id_column: Name of column containing job ids
    :type job_id_column: str
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    """
    logger.info("Computing missing values")

    job_names = data[job_id_column].unique()

    for job_name in job_names:
        per_job_df = data[data[job_id_column].str.contains(job_name)] 
        per_job_df = per_job_df.drop(job_id_column, axis=1)
        missing_data = per_job_df.isnull().sum()
        missing_percentage = (missing_data / len(per_job_df)) * 100
        missing_df = pd.DataFrame({"Missing Values": missing_data,
                                  "Percentage": missing_percentage})
        missing_df = missing_df.sort_values(
                by="Percentage", ascending=False)
        
        if len(missing_df[missing_df["Missing Values"] > 0]) > 0:
            output_file_missing = f"missing-values-{job_name}.csv"
            written_file = utils.write_df_to_csv(
                missing_df, output_path, output_file_missing)
            logger.info((f"Missing value stats for {job_name}" 
                        f"saved to {written_file}"))
        else:
            logger.info(f"No missing values found for {job_name}.")
   

def _summarize_data(data: pd.DataFrame, job_id_column: str, 
                    output_path: str) -> None:
    """
    Computes summary statistics and writes results to a CSV file.

    :param data: Data execution history data preprocessed from
    measurements and prepared by _prepare_data.
    :type data: pandas.DataFrame
    :param job_id_column: Name of column containing job ids
    :type job_id_column: str
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    """
    logger.info("Summarizing data")

    output_file_summary = "descriptive-stats.csv"
    summary_df = data.groupby(job_id_column).describe()
    written_file = utils.write_df_to_csv(
        summary_df, output_path, output_file_summary)
    logger.info(f"Summary statistics written to {written_file}.")
    

def _correlate_data(data: pd.DataFrame, job_id_column: str,
                    output_path: str) -> None:
    """
    Computes pairwise correlations and writes to a CSV file.

    :param data: Data execution history data preprocessed from
    measurements and prepared by _prepare_data.
    :type data: pandas.DataFrame
    :param job_id_column: Name of column containing job ids
    :type job_id_column: str
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    """
    logger.info("Computing correlations")

    job_names = data[job_id_column].unique()

    for job_name in job_names:
        corr_df = data[data[job_id_column].str.contains(job_name)]
        corr_df = corr_df.drop(job_id_column, axis=1)
        corr_df = corr_df.select_dtypes(include=constants.NUMERICS)
        corr_df = corr_df.corr(method="pearson")
        output_file_corr = f"correlations-{job_name}.csv"
        written_file = utils.write_df_to_csv(
            corr_df, output_path, output_file_corr)
        logger.info(f"Correlations for {job_name} saved to {written_file}")


def _summarize_categorical_data(data: pd.DataFrame, 
                                target_variables: list[str],
                                output_path: str) -> None:
    """
    Computes aggregations of each target variable per value of a
    categorical variable.
    :param data: Data execution history data preprocessed from
    measurements and prepared by _prepare_data. 
    :type raw_data: pandas.DataFrame
    :param target_variables: List of column names containing target 
    variables (from job_spec.yaml)
    :type target_variables: list[str]
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    """
    logger.info(("Computing target variable aggregations for categorical" 
                " variables"))

    categorical_variables: list[str] = utils.get_categorical_features(data)
    logger.info(f"Found categorical variables: {categorical_variables}")

    # We remove start_time and end_time fields and instead rely on 
    # a duration field being present. 
    # TODO discuss a better long-term approach
    target_variables = [x for x in target_variables 
                        if not x.endswith("start_time") 
                        or not x.endswith("end_time")]
    target_variables.insert(0, constants.JOB_DURATION_FIELD_NAME)
    
    # We remove target variables whose columns have no data
    target_variables = [x for x in target_variables if x in data.columns]
    logger.info(f"Analyzing for target variables: {target_variables}")

    for target_var in target_variables:
        agg_df = data.groupby(categorical_variables).agg(
            target_var_min=(target_var, "min"),
            target_var_mean=(target_var, "mean"),
            target_var_max=(target_var, "max"))
        output_file_categorical = f"categorical-stats-{target_var}.csv"
        written_file = utils.write_df_to_csv(
            agg_df, output_path, output_file_categorical)
        logger.info(f"Categorical summary written to {written_file}")


def _plot_histograms(data: pd.DataFrame,
                     job_id_column: str,
                     output_path: str) -> None:
    """
    Plots histograms of all numeric variables and stores them to a single 
    PDF file.

    :param data: Data execution history data preprocessed from
    measurements and prepared by _prepare_data.
    :type data: pandas.DataFrame
    :param job_id_column: Name of column containing job ids
    :type job_id_column: str
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    """
    logger.info("Generating histograms")
    
    if not os.path.exists(output_path):
        os.makedirs(output_path) 

    nrows = 3
    ncols = 1

    for job_name, data_slice in data.groupby(job_id_column):
        output_file = os.path.join(
            output_path, f"histogram-{job_name}.pdf")
        pdf = PdfPages(output_file)
        data_slice = data_slice.drop(job_id_column, axis=1)
        data_slice = data_slice.select_dtypes(include=constants.NUMERICS)
        cols = data_slice.columns

        for i, col in enumerate(cols):
            j = i % (nrows * ncols)

            if j == 0:
                fig = plt.figure(figsize=(8, 6))
            
            ax = fig.add_subplot(nrows, ncols, j + 1)
            ax.set_title(f"{job_name} - {col}")
            data_slice[col].plot(kind="hist", grid=True, ax=ax, bins=15)
    
            if j == (nrows * ncols) - 1 or i == len(cols) - 1:
                fig.tight_layout()
                pdf.savefig(fig)
                logger.info(
                    f"Histograms for {job_name} saved to {output_file}")
        pdf.close()
    

def _plot_box_whiskers(data: pd.DataFrame, job_id_column: str,
                       output_path: str) -> None:
    """
    Plots box-whisper plots and stores them to a single PDF file. 

    :param data: Data execution history data preprocessed from
    measurements and prepared by _prepare_data.
    :type data: pandas.DataFrame
    :param job_id_column: Name of column containing job ids
    :type job_id_column: str
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    """
    logger.info("Generating box-whisker plots")

    if not os.path.exists(output_path):
        os.makedirs(output_path) 

    nrows = 1
    ncols = 4

    for job_name, data_slice in data.groupby(job_id_column):
        output_file = os.path.join(output_path, f"box-{job_name}.pdf")
        pdf = PdfPages(output_file)
        data_slice = data_slice.drop(job_id_column, axis=1)
        data_slice = data_slice.select_dtypes(include=constants.NUMERICS)
        cols = data_slice.columns

        for i, col in enumerate(cols):
            j = i % (nrows * ncols)

            if j == 0:
                fig = plt.figure(figsize=(12, 5))
            
            ax = fig.add_subplot(nrows, ncols, j + 1)
            ax.set_title(f"{job_name} - {col}")
            data_slice[col].plot(kind="box", grid=True, ax=ax)
    
            if j == (nrows * ncols) - 1 or i == len(cols) - 1:
                fig.tight_layout()
                pdf.savefig(fig)
                logger.info(
                    f"Box plot for {job_name} saved to {output_file}")
        pdf.close()
        

def _plot_correlations(data: pd.DataFrame, job_id_column: str,
                       output_path: str) -> None:
    """
    Plots correlation matrices for each job.

    :param data: Data execution history data preprocessed from
    measurements and prepared by _prepare_data.
    :type data: pandas.DataFrame
    :param job_id_column: Name of column containing job ids
    :type job_id_column: str
    :param output_path: Path to directory to use for data analysis files.
    :type output_path: str
    """
    logger.info("Generating correlation matrix plots")

    if not os.path.exists(output_path):
        os.makedirs(output_path) 

    fig, ax = plt.subplots()

    for job_name, data_slice in data.groupby(job_id_column):
        data_slice = data_slice.drop(job_id_column, axis=1)
        data_slice = data_slice.select_dtypes(include=constants.NUMERICS)
        fig = plt.figure(figsize=(14, 14))
        ax = fig.add_subplot(111)
        cax = ax.matshow(data_slice.corr(), vmin=-1, vmax=1)
        ticks = np.arange(0, len(data_slice.columns), 1)
        ax.set_xticks(ticks)
        ax.set_yticks(ticks)
        ax.set_xticklabels(data_slice.columns, rotation=45)
        ax.set_yticklabels(data_slice.columns, rotation=45)
        fig.colorbar(cax)
        plt.title(f"{job_name}")
        output_file = os.path.join(
            output_path, f"correlation-{job_name}.png")
        plt.savefig(output_file, dpi=300)


# TODO complete implementation and test
def analyze_job_data(raw_data: pd.DataFrame, job_id_column: str = None,
                     custom_job_name: str = None, output_path: str = None,
                     target_variables: list[str] = None):
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    
    logger.info("Beginning descriptive job analysis.")

    job_id_column, custom_job_name = _init_args(
        job_id_column=job_id_column, custom_job_name=custom_job_name, 
        output_path=output_path)
    prepared_data = _prepare_data(data=raw_data, job_id_column=job_id_column,
                                  custom_job_name=custom_job_name)
    _compute_missing_data(data=prepared_data, job_id_column=job_id_column,
                          output_path=output_path) 
    _summarize_data(data=prepared_data, job_id_column=job_id_column, 
                    output_path=output_path)
    _correlate_data(data=prepared_data, job_id_column=job_id_column, 
                    output_path=output_path)
    _summarize_categorical_data(data=prepared_data, 
                                target_variables=target_variables, 
                                output_path=output_path)
    _plot_histograms(data=prepared_data, job_id_column=job_id_column, 
                     output_path=output_path)
    _plot_box_whiskers(data=prepared_data, job_id_column=job_id_column, 
                       output_path=output_path)
    _plot_correlations(data=prepared_data, job_id_column=job_id_column,
                       output_path=output_path)
    logger.info(f"Wrote descriptive job analysis to {output_path}.")



import numpy as np
from typing import Tuple

from sklearn.metrics import (
    make_scorer,
    mean_squared_error, 
    mean_absolute_percentage_error, 
    r2_score
    )
                             
"""
Custom ML performance metrics and related utilities.
"""


def normalized_root_mean_squared_error_minmax(y_true, y_pred):
    rmse_mean = np.sqrt(mean_squared_error(y_true, y_pred))
    target_variable_min, target_variable_max = get_min_max_from_array(y_true)
    nrmse = rmse_mean / (target_variable_max - target_variable_min)
    return nrmse


def compute_test_metrics(y_true, y_pred):
    mape_mean_test_all = [mean_absolute_percentage_error([y_t], [y_p]) for y_t, y_p in zip(y_true, y_pred)]
    mape_mean_test = np.average(mape_mean_test_all)
    mape_std_test = np.std(mape_mean_test_all)
    nrmse_mean_test = normalized_root_mean_squared_error_minmax(y_true, y_pred)
    r2_mean_test = r2_score(y_true, y_pred)
    return mape_mean_test, mape_std_test, nrmse_mean_test, r2_mean_test


def create_scorers():
    scoring = {
        "neg_mean_absolute_percentage_error": 
            "neg_mean_absolute_percentage_error", 
        "neg_normalized_root_mean_squared_error(minmax)": make_scorer(
            normalized_root_mean_squared_error_minmax, 
            greater_is_better=False), 
        "neg_root_mean_squared_error": "neg_root_mean_squared_error", 
        "neg_mean_squared_error": "neg_mean_squared_error", 
        "neg_mean_absolute_error": "neg_mean_absolute_error", 
        "r2": "r2"
    }
    return scoring


def get_min_max_from_array(array) -> Tuple[float, float]:
    min = np.amin(array)
    max = np.amax(array)
    return min, max



import shutil

import pandas as pd
import os
from arise_predictions.cmd.cmd import parse_args, get_args
from arise_predictions.job_statistics.analyze_jobs import analyze_job_data
from arise_predictions.auto_model.build_models import auto_build_models, get_estimators_config
from arise_predictions.perform_predict.predict import demo_predict, data_predict, get_predict_config
from arise_predictions.utils import constants, utils
from arise_predictions.preprocessing import job_parser
import logging

logger = logging.getLogger(__name__)


def load_spec(spec_path):
    # analyzing job spec file

    spec_folder = utils.get_unpacked_path(spec_path)

    job_spec_file = os.path.join(spec_folder, constants.JOB_SPEC_FILE_NAME)
    logger.info('Analyzing job spec file: %s', job_spec_file)
    loaded_job_spec = job_parser.parse_job_spec(job_spec_file)
    if not loaded_job_spec:
        logger.error("Failed to load job spec")
        raise Exception
    return loaded_job_spec


def execute_preprocess(spec_path):

    job_spec = load_spec(spec_path)

    inputs = sorted(list(job_spec[0]))
    outputs = sorted(list(job_spec[1]))
    start_time_field_name = job_spec[2]
    end_time_field_name = job_spec[3]
    job_parser_class_name = job_spec[4]
    job_entry_filter = job_spec[5]
    feature_engineering = job_spec[6] if len(job_spec) > 6 else None
    metadata_parser_class_name = job_spec[7] if len(job_spec) > 7 else None

    # processing history ( if not done in the past )
    analyzed_history_file = os.path.join(
        get_args().input_path, constants.JOB_HISTORY_FILE_NAME + ".csv")
    return get_history(analyzed_history_file, inputs, outputs, start_time_field_name, end_time_field_name,
                       job_parser_class_name, job_entry_filter, feature_engineering, metadata_parser_class_name)


def execute_analyze_jobs():

    # processing history ( if not done in the past )
    history_data, history_file = execute_preprocess(get_args().input_path)

    if history_data is None or history_data.empty:
        logging.error(("No historical data could be retrieved from given" 
                       " location {}").format(get_args().input_path))
    else:
        loaded_job_spec = load_spec(get_args().input_path)
        outputs = sorted(list(loaded_job_spec[1]))
        logging.info("Invoking job analysis")
        analyze_job_data(raw_data=history_data, job_id_column=get_args().job_id_column,
                         custom_job_name=get_args().custom_job_name,
                         output_path=os.path.join(get_args().input_path,
                                                  constants.JOB_ANALYSIS_PATH),
                         target_variables=outputs)


def execute_auto_build_models():

    # processing history ( if not done in the past )
    history_data, history_file = execute_preprocess(get_args().input_path)

    if history_data is None or history_data.empty:
        logging.error(("No historical data could be retrieved from given" 
                       " location {}").format(get_args().input_path))
    else:
        # First copy job spec file and metadata dir (if exists) to output path,
        # so they are available later for prediction. This is done before building the models,
        # because afterwards output folder may be zipped and deleted.
        output_path = os.path.join(get_args().input_path, constants.AM_OUTPUT_PATH_SUFFIX)
        utils.mkdirs(output_path)
        shutil.copy(os.path.join(get_args().input_path, constants.JOB_SPEC_FILE_NAME), output_path)
        loaded_job_spec = load_spec(get_args().input_path)
        outputs = sorted(list(loaded_job_spec[1]))
        feature_engineering = loaded_job_spec[6] if len(loaded_job_spec) > 6 else None
        if feature_engineering and not get_args().ignore_metadata:
            shutil.copytree(os.path.join(get_args().input_path, constants.JOB_METADATA_DIR),
                            os.path.join(output_path, constants.JOB_METADATA_DIR),
                            dirs_exist_ok=True)
        logging.info("Invoking auto model search and build")
        auto_build_models(raw_data=history_data,
                          config=get_estimators_config(config_file=get_args().config_file,
                                                       num_jobs=get_args().num_jobs),
                          target_variables=outputs,
                          output_path=output_path,
                          leave_one_out_cv=get_args().leave_one_out_cv,
                          feature_col=get_args().feature_column,
                          low_threshold=get_args().low_threshold,
                          high_threshold=get_args().high_threshold,
                          single_output_file=get_args().single_output_file,
                          randomized_hpo=get_args().randomized_hpo,
                          n_random_iter=get_args().random_iterations)


def execute_demo_predict():

    # processing history ( if not done in the past )
    history_data, history_file = execute_preprocess(get_args().model_path)

    if history_data is None or history_data.empty:
        logging.error(("No historical data could be retrieved from given" 
                       " location {}").format(get_args().input_path))
    else:
        loaded_job_spec = load_spec(get_args().model_path)
        logging.info("Invoking demo predict")
        demo_predict(
            original_data=history_data,
            config=get_predict_config(get_args().config_file),
            estimator_path=get_args().model_path,
            feature_engineering=None if get_args().ignore_metadata else loaded_job_spec[6],
            metadata_parser_class_name=loaded_job_spec[7],
            metadata_path=get_args().model_path,
            output_path=os.path.join(get_args().input_path, constants.PRED_OUTPUT_PATH_SUFFIX))


def execute_predict():
    loaded_job_spec = load_spec(get_args().model_path)

    logging.info("Invoking predict")
    demo_predict(
        original_data=None,
        config=get_predict_config(get_args().config_file),
        estimator_path=get_args().model_path,
        feature_engineering=None if get_args().ignore_metadata else loaded_job_spec[6],
        metadata_parser_class_name=loaded_job_spec[7],
        metadata_path=get_args().model_path,
        output_path=os.path.join(get_args().input_path, constants.PRED_OUTPUT_PATH_SUFFIX))


def execute_data_predict():
    loaded_job_spec = load_spec(get_args().model_path)
    input_path = get_args().input_path

    logging.info("Invoking data predict")
    data_predict(
        original_data=pd.read_csv(os.path.join(input_path, get_args().original_data_file)) if
        get_args().original_data_file else None,
        prediction_data=pd.read_csv(os.path.join(input_path, get_args().prediction_data_file)),
        estimator_path=get_args().model_path,
        estimators_config=get_predict_config(get_args().config_file).estimators,
        target_variables=loaded_job_spec[1],
        delta_only=get_args().delta_only,
        output_path=os.path.join(input_path, constants.PRED_OUTPUT_PATH_SUFFIX))


def get_history(history_file, inputs, outputs, start_time_field_name, end_time_field_name, job_parser_class_name,
                job_entry_filter, feature_engineering, metadata_parser_class_name):
    if os.path.exists(history_file) and not get_args().reread_history:
        logging.info("using already processed history")
        history_data = pd.read_csv(history_file)
        history_data = history_data[utils.adjust_columns_with_duration(history_data.columns.values.tolist(),
                                                                       start_time_field_name,
                                                                       end_time_field_name)]
    else:
        logging.info("processing historical jobs")
        history_data, history_file = job_parser.collect_jobs_history(
            os.path.join(get_args().input_path, constants.JOB_DATA_DIR), get_args().input_path, inputs, outputs,
            start_time_field_name, end_time_field_name, get_args().input_file, job_parser_class_name, job_entry_filter,
            None if get_args().ignore_metadata else feature_engineering, metadata_parser_class_name,
            get_args().input_path)
    return history_data, history_file


def main():
    if not parse_args():
        global logger
        logger.error('Failed to parse command line arguments')
        exit(1)
    level = logging.getLevelName(get_args().loglevel.upper())
    print("level: %d" % level)
    print("cmd_args: %s" % get_args())

    logger = logging.getLogger()
    logging.basicConfig(
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=level)

    # According to the selected command, call the appropriate function
    if get_args().command == 'preprocess':
        execute_preprocess(get_args().input_path)
    elif get_args().command == 'analyze-jobs':
        execute_analyze_jobs()
    elif get_args().command == 'auto-build-models':
        execute_auto_build_models()
    elif get_args().command == 'demo-predict':
        execute_demo_predict()
    elif get_args().command == 'predict':
        execute_predict()
    elif get_args().command == 'data-predict':
        execute_data_predict()
    else:
        logger.error('Invalid command!')
        logger.info(
            'For development purpose we will execute the get-stats command')
        logger.info('This will be removed in production')
        from arise_predictions.cmd.cmd import cmd_args
        cmd_args.command = "get-stats"
        cmd_args.input_path = "examples/MLCommons"
        cmd_args.reread_history = False
        execute_analyze_jobs()


if __name__ == '__main__':
    main()



JOB_INPUTS_FIELD_NAME = 'job-metadata-inputs'
JOB_OUTPUTS_FIELD_NAME = 'job-metadata-outputs'
JOB_START_TIME_FIELD_NAME = 'start-time-field-name'
JOB_END_TIME_FIELD_NAME = 'end-time-field-name'
JOB_START_TIME_DEFAULT_FIELD_NAME = 'start_time'
JOB_END_TIME_DEFAULT_FIELD_NAME = 'end_time'
JOB_DURATION_FIELD_NAME = 'duration'
JOB_HISTORY_FILE_NAME = "data_execution_history"
JOB_METADATA_FILE_NAME = "data_execution_metadata"
DEFAULT_CORR_THRESHOLD = 0.85
KS_TEST_SIGNIFICANCE_LEVEL = 0.05
JOB_PARSER_CLASS_NAME_FIELD = 'job-parser-class-name'
METADATA_PARSER_CLASS_NAME_FIELD = 'metadata-parser-class-name'
JOB_ENTRY_FILTER_FIELD = 'job-entry-filter'
JOB_ENTRY_FILTER_NAME_COL = 'name'
JOB_ENTRY_FILTER_VALUES_COL = 'excluded_values'
JOB_ENTRY_FILTER_KEEP_COL = 'keep_input'
DUMMY_VARS_PREFIX = 'dummy_input_'
JOB_INPUTS_FEATURE_ENGINEERING = 'job-metadata-fe'
JOB_DATA_DIR = "data"
JOB_METADATA_DIR = "metadata"
JOB_SPEC_FILE_NAME = "job_spec.yaml"


# job data analysis
JOB_ANALYSIS_PATH = "job-analysis"
DEFAULT_JOB_ID_COLUMN = "job_name"
DEFAULT_JOB_NAME = "single-job"
NUMERICS = ["int16", "int32", "int64", "float16", "float32", "float64"]

# matplotlib
FONT_SMALL_SIZE = 8
FONT_MEDIUM_SIZE = 10

# auto model
SEED = 42
AM_CONFIG_ESTIMATORS = "estimators"
AM_CONFIG_NAME = "name"
AM_CONFIG_CLASS_NAME = "class_name"
AM_CONFIG_LINEAR = "linear"
AM_CONFIG_PARAMETERS = "parameters"
AM_ESTIMATORS_NO_SEED = ["LinearRegression", "SVR", "DecisionTreeRegressor",
                         "KNeighborsRegressor"]
AM_ESTIMATORS_CATBOOST = ["CatBoostRegressor"]
AM_DEFAULT_METRIC = "neg_mean_absolute_percentage_error"
AM_OUTPUT_PATH_SUFFIX = "ARISE-auto-models"
AM_NUM_JOBS = "num_jobs"
AM_RANKINGS_FILE = "all-star-rankings.csv"
AM_META_LEARNER_PREFIX = "meta-learner-"
AM_DEFAULT_N_ITER_RANDOM_HPO = 50

# auto model column headers
AM_COL_ESTIMATOR = "estimator"
AM_COL_LINEAR = "linear"
AM_COL_TARGET = "target_variable"
AM_COL_TARGET_RANGE = "target_variable_range"
AM_COL_RANK_MAPE = "rank_MAPE"
AM_COL_RANK_NRMSE_MAXMIN = "rank_NRMSE_maxmin"
AM_COL_RANK_R2 = "rank_R2"

# auto model performance metrics
# TODO consider moving to metrics.py
AM_METRIC_MAPE_MEAN = "mean_test_neg_mean_absolute_percentage_error"
AM_METRIC_NRMSE_MEAN = "mean_test_neg_normalized_root_mean_squared_error(minmax)"
AM_METRIC_R2_MEAN = "mean_test_r2"

# predict
PRED_OUTPUT_PATH_SUFFIX = "ARISE-predictions"
PRED_INPUT_SPACE_FILE = "input-space.csv"
PRED_ALL_PREDICTIONS_FILE = "all-predictions.csv"
PRED_GROUND_TRUTH_FILE = "predictions-with-ground-truth.csv"
PRED_ORIGINAL_TRUTH_FILE = "original-data.csv"
PRED_CONFIG_FIXED = "fixed_values"
PRED_CONFIG_VARIABLE = "variable_values"
PRED_CONFIG_DATA = "data_values"
PRED_CONFIG_DATA_INPUT = "input_variable"
PRED_CONFIG_DATA_VALUES = "values"
PRED_CONFIG_DATA_ALL = "all"
PRED_CONFIG_DATA_MIN_MAX = "min_max"
PRED_CONFIG_DATA_EXCLUDE = "exclude"
PRED_CONFIG_INTERPOLATION = "interpolation_values"
PRED_CONFIG_ESTIMATORS = "estimators"
PRED_CONFIG_TARGET_VAR = "target_variable"
PRED_CONFIG_GREATER_BETTER = "greater_is_better"
PRED_CONFIG_ESTIMATOR_FILE = "estimator_file"

# extrapolation
EXTRA_TEST_FILE = "extrapolation-test-data.csv"



import unittest
import logging
import sys
import os

import pandas as pd

from tests.utils.logger_redirector import LoggerRedirector
from arise_predictions.utils import constants
from arise_predictions.perform_predict.predict import (_create_input_space, _run_predictions, get_predict_config)

"""
Tests for perform_predict.predict.
"""

logger = logging.getLogger('test-logger')

logging.basicConfig(
    format="%(asctime)s %(module)s %(levelname)s: %(message)s",
    level=logging.INFO,
    stream=sys.stdout)


class TestPredict(unittest.TestCase):

    def setUp(self):
        # unittest has reassigned sys.stdout and sys.stderr by this point
        LoggerRedirector.redirect_loggers(fake_stdout=sys.stdout, 
                                          fake_stderr=sys.stderr) 
        
        self.resources_path = "tests/resources"
        self.metadata_path = "tests/resources"
        self.predict_config_file = os.path.join(self.resources_path, 
                                              "example-demo-mlcommons-config.yaml")
        self.script_dir = os.path.dirname(__file__)
        self.output_path = os.path.abspath(os.path.join(
               self.script_dir, "test-tmp", constants.PRED_OUTPUT_PATH_SUFFIX))
        self.input_space_df_file = "input-space.csv"

        self.predictions_thpt_file = "predictions-tokens_per_second.csv"
        self.predictions_all_file = constants.PRED_ALL_PREDICTIONS_FILE

        self.data_execution_history_file = "mlcommons_inference_data_execution_history.csv"
    
    def tearDown(self):
        LoggerRedirector.reset_loggers(
            fake_stdout=sys.stdout, fake_stderr=sys.stderr)
    
        # if os.path.exists(self.output_path):
        #     shutil.rmtree(self.output_path)

    def test_read_input_config_values(self):
        """
        Reads input feature space configuration as expected.
        """
        input_config = get_predict_config(self.predict_config_file)
        logger.info(f"fixed_values {input_config.fixed_values}")
        logger.info(f"variable_values {input_config.variable_values}")
        self.assertIsNotNone(input_config.fixed_values)
        self.assertIsNotNone(input_config.variable_values)
        self.assertEqual(7, len(input_config.fixed_values))
        self.assertEqual(3, len(input_config.variable_values))

    def test_create_input_space(self):
        """
        DataFrame representing input space is created.
        """
        input_config = get_predict_config(self.predict_config_file)
        original_data_df = pd.read_csv(os.path.join(
            self.resources_path,
            self.data_execution_history_file))
        path, _ = _create_input_space(input_config,  original_data_df, self.output_path)
        self.assertEqual(path, os.path.join(
            self.output_path,
            self.input_space_df_file
        ))
        self.assertTrue(os.path.exists(path))

    def test_read_estimator_config(self):
        """
        Reads estimator configuration for each target variable.
        """
        estimator_config = get_predict_config(self.predict_config_file)
        logger.info(f"estimator_config: {estimator_config}")
        self.assertTrue(2, len(estimator_config.estimators))

    def test_predictions_exist(self):
        """
        Running predictions produces CSV files as expected.

        Note that we are using an input space generated from inference latency
        data using the test functions above and estimator binaries created by
        auto-model build on the same data.
        """
        config = get_predict_config(self.predict_config_file)
        original_data_df = pd.read_csv(os.path.join(
            self.resources_path, 
            self.data_execution_history_file))

        _, input_space_df = _create_input_space(
            input_config=config,
            original_data= original_data_df,
            metadata_path=self.metadata_path,
            feature_engineering=None,
            output_path=self.output_path
        ) 
        
        _run_predictions(
            original_data=original_data_df,
            input_data=input_space_df,
            estimators_config=config.estimators,
            estimator_path=self.resources_path,
            output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.predictions_thpt_file)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, self.predictions_all_file)))
        predictions_all_df = pd.read_csv(os.path.join(
            self.output_path, self.predictions_all_file
        ))
        self.assertIn("rank_tokens_per_second", predictions_all_df.columns)
        


import sys
import logging


class LoggerRedirector:

    """
    Trickush to get unittest to display logs of code failing tests. See
    https://stackoverflow.com/questions/69200881/how-to-get-python-unittest-to-show-log-messages-only-on-failed-tests
    for more details.
    """

    # Keep a reference to the real streams so we can revert
    _real_stdout = sys.stdout
    _real_stderr = sys.stderr

    @staticmethod
    def all_loggers():
        loggers = [logging.getLogger()]
        loggers += [logging.getLogger(name) for name in logging.root.manager.loggerDict]
        return loggers

    @classmethod
    def redirect_loggers(cls, fake_stdout=None, fake_stderr=None):
        if ((not fake_stdout or fake_stdout is cls._real_stdout)
                and (not fake_stderr or fake_stderr is cls._real_stderr)):
            return
        for logger in cls.all_loggers():
            for handler in logger.handlers:
                if hasattr(handler, 'stream'):
                    if handler.stream is cls._real_stdout:
                        handler.setStream(fake_stdout)
                    if handler.stream is cls._real_stderr:
                        handler.setStream(fake_stderr)

    @classmethod
    def reset_loggers(cls, fake_stdout=None, fake_stderr=None):
        if ((not fake_stdout or fake_stdout is cls._real_stdout)
                and (not fake_stderr or fake_stderr is cls._real_stderr)):
            return
        for logger in cls.all_loggers():
            for handler in logger.handlers:
                if hasattr(handler, 'stream'):
                    if handler.stream is fake_stdout:
                        handler.setStream(cls._real_stdout)
                    if handler.stream is fake_stderr:
                        handler.setStream(cls._real_stderr)


import unittest
import os
import sys
import logging

import pandas as pd

from arise_predictions.utils import utils, constants
from tests.utils.logger_redirector import LoggerRedirector

from arise_predictions.auto_model.build_models import (
    get_estimators_config,
    _init_estimators,
    _search_models,
    _rank_estimators,
    _select_best_estimators,
    _search_meta_model,
    _merge_rankings,
    _persist_and_test_meta_estimator
)

from arise_predictions.cmd.cmd import parse_args

"""
Tests for building of meta-learner (for improved extrapolation performance).

Note that the tearDown method deletes any temporary test files that may be
useful for debugging. Have left shutil.rmtree commented out to avoid this.
"""

logger = logging.getLogger('test-logger')

logging.basicConfig(
    format="%(asctime)s %(module)s %(levelname)s: %(message)s",
    level=logging.INFO,
    stream=sys.stdout)


class TestBuildMetaLearner(unittest.TestCase):

    def setUp(self):
        LoggerRedirector.redirect_loggers(fake_stdout=sys.stdout, 
                                          fake_stderr=sys.stderr) 

        # Simulate command-line arguments
        parse_args(["--num-jobs", "1"])

        self.resources_path = "tests/resources"
        self.config_file = os.path.join(self.resources_path, "model-search-config.yaml")
        self.data_file = "mlcommons_inference_data_execution_history.csv"
        self.target_variables = ["tokens_per_second"]

        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        self.categorical_variables: list[str] = utils.get_categorical_features(data)
        self.categorical_variables_indices = [data.columns.get_loc(var) for var in self.categorical_variables]

        self.script_dir = os.path.dirname(__file__)
        self.output_path = os.path.abspath(os.path.join(
               self.script_dir, "test-tmp/ARISE-auto-models"))

        self.feature_col = "# of Accelerators"
        self.high_threshold = 6
        self.extrapolation_file = "extrapolation-test-data.csv"

        self.best_estimator_name_linear_gpu = "RidgeRegression"
        self.best_estimator_name_nonlinear_gpu = "XGBoost-Regressor"
        self.best_estimator_name_linear_time = "RidgeRegression"
        self.best_estimator_name_nonlinear_time = "XGBoost-Regressor"

        self.estimator_name_meta = constants.AM_META_LEARNER_PREFIX + "ElasticNet-Regression"

        self.results_meta_learner_en = ("cv_results-meta-learner-"
                                             "ElasticNet-Regression-"
                                             "tokens_per_second.csv")
        
        self.meta_rankings_file = constants.AM_META_LEARNER_PREFIX + constants.AM_RANKINGS_FILE
        self.all_rankings_file = constants.AM_RANKINGS_FILE

        self.estimator_file_name_meta_gpu = ("estimator-linear-meta-learner-"
                                             "ElasticNet-Regression-"
                                             "gpu_memory_utilization_max.pkl")
        self.estimator_file_name_meta_time = ("estimator-linear-meta-learner-"
                                              "ElasticNet-Regression-"
                                              "train_runtime.pkl")
        self.estimator_meta_test_file = ("best-meta-estimators-testset-"
                                         "performance-summary.csv")
        self.estimator_meta_extrapolation_test_file = ("best-meta-estimators-"
                                                       "extrapolation-testset-"
                                                       "performance-summary.csv")

    def tearDown(self):
        LoggerRedirector.reset_loggers(
            fake_stdout=sys.stdout, fake_stderr=sys.stderr)
        
        # if os.path.exists(self.output_path):
        #     shutil.rmtree(self.output_path)
    
    def test_meta_learner(self):
        """
        Going against best practices and having a single method exercising all
        functionality I want to test (via separate assertion methods) to avoid
        having to run auto-model-build repeatedly (time-consuming).
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
                config, cat_indices=self.categorical_variables_indices)
        self.assertEqual(len(estimators), 4, "4 estimators")
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))

        stats_df, estimators_per_target_variable = _search_models(
            data=data, estimators=estimators, 
            config=config,
            categorical_variables=self.categorical_variables,
            target_variables=self.target_variables, 
            output_path=self.output_path,
            feature_col=self.feature_col,
            high_threshold=self.high_threshold)
        
        self._extrapolation_test_data_exists_if_feature_col()

        rankings = _rank_estimators(
            summary_stats=stats_df, 
            output_path=self.output_path,
            output_file=constants.AM_RANKINGS_FILE)

        best_estimators_for_target_variables = _select_best_estimators(
            target_variables=self.target_variables,
            rankings=rankings,
            estimators_per_target_variable=estimators_per_target_variable,
            output_path=self.output_path,
            num_jobs=config.num_jobs,
            leave_one_out_cv=None,
            categorical_variables=self.categorical_variables)
        
        self._best_estimators_as_expected(
            best_estimators_for_target_vars=best_estimators_for_target_variables)
        
        stats_df, best_meta_learner_per_target_variable = _search_meta_model(
            best_estimators_per_target_var=best_estimators_for_target_variables,
            config=config,
            cat_indices=self.categorical_variables_indices,
            target_variables=self.target_variables,
            output_path=self.output_path
        )

        self._meta_learner_results_as_expected(
            stats_df=stats_df, 
            best_meta_learner_per_target_variable=best_meta_learner_per_target_variable)
        
        _ = _rank_estimators(
            summary_stats=stats_df,
            output_path=self.output_path,
            output_file=self.meta_rankings_file)
        
        _merge_rankings(output_path=self.output_path) 
        
        self._meta_learner_rankings_exists()

        _persist_and_test_meta_estimator(
            meta_learner_per_target_variable=best_meta_learner_per_target_variable,
            output_path=self.output_path,
            num_jobs=config.num_jobs,
            leave_one_out_cv=None,
            categorical_variables=self.categorical_variables
        )

    def _extrapolation_test_data_exists_if_feature_col(self):
        """
        Extrapolation test data exists if feature column has been set.
        """
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.extrapolation_file)))

        df = pd.read_csv(os.path.join(
            self.output_path,
            self.extrapolation_file
        ))
        self.assertGreater(len(df), 0, 
                           "Extrapolation DF should contain at least 1 row.")
        self.assertTrue(all(df[self.feature_col] >= self.high_threshold),
                        (f"All values in {self.feature_col} should be at" 
                        f" least {self.high_threshold}"))
    
    def _best_estimators_as_expected(self, best_estimators_for_target_vars):
        """
        List of best estimators is as expected.
        """
        self.assertEqual(len(best_estimators_for_target_vars), 1,
                         "One target variable")
        best_estimators_for_thpt = best_estimators_for_target_vars[
            self.target_variables[0]]
        self.assertEqual(len(best_estimators_for_thpt), 2,
                         "2 best estimators per target variable")
        estimator_names_thpt = [estimator["estimator_name"] for estimator in best_estimators_for_thpt]
        self.assertIn(self.best_estimator_name_linear_gpu, estimator_names_thpt)
        self.assertIn(self.best_estimator_name_nonlinear_gpu, estimator_names_thpt)

    def _meta_learner_results_as_expected(self, stats_df, best_meta_learner_per_target_variable):
        """
        Meta learner CV results exist and have one meta-learner per target variable.
        """
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.results_meta_learner_en
        )))
        self.assertEquals(
            len(best_meta_learner_per_target_variable[self.target_variables[0]]),
            1, 
            "1 meta-learner per target variable"
        )
        utils.write_df_to_csv(stats_df, self.output_path, "A_STATS_DF_INSPECT.csv")

    def _meta_learner_rankings_exists(self):
        """
        The meta-all-star-rankings exists and is data is in all-star-rankings.
        """
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.meta_rankings_file 
        )))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.all_rankings_file 
        ))) 
        rankings_df = pd.read_csv(os.path.join(
            self.output_path, 
            self.all_rankings_file
            )
        )
        self.assertIn(
            self.estimator_name_meta, 
            rankings_df[constants.AM_COL_ESTIMATOR].tolist())

    def _meta_learner_persisted_and_tested(self):
        """
        Meta-learner has been persisted along with test results.
        """
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.estimator_file_name_meta_gpu
        )))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.estimator_file_name_meta_time
        )))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.estimator_meta_test_file
        )))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.estimator_meta_extrapolation_test_file
        )))




import yaml
import os
import pandas as pd
import numpy as np

from arise_predictions.main import logger
from arise_predictions.utils import utils, constants
from arise_predictions.preprocessing import custom_job_parser
import logging
from typing import List

logger = logging.getLogger(__name__)

"""
This module parsers a job spec and collects its historical executions from different representations
into CSV format based on the given spec. All spec fields are assumed to be numeric.
"""


def parse_job_spec(job_spec_file):
    """ Parse job_spec_file.
    If this variable is a string, assume it is a file name. Then, open and load spec from file.
    Else, consider this as the job spec dict.
    """

    if isinstance(job_spec_file, dict):
        loaded_spec = job_spec_file
    elif isinstance(job_spec_file, str):
        # Reading the job spec file
        with open(job_spec_file, "r") as file:
            try:
                loaded_spec = dict(yaml.safe_load(file))
            except yaml.YAMLError as exc:
                logger.error(exc)
                return None
    else:
        logger.error("job_spec_file is neither a dictionary nor a filename string")
        return None

    # Making sure that all required fields are available in the spec
    for field in [constants.JOB_INPUTS_FIELD_NAME, constants.JOB_OUTPUTS_FIELD_NAME]:
        if field not in loaded_spec:
            logger.error("Missing key: {}".format(field))
            return None

    # Reading start and end time field names, using defaults if not provided
    if constants.JOB_START_TIME_FIELD_NAME in loaded_spec:
        start_time_field_name = loaded_spec[constants.JOB_START_TIME_FIELD_NAME]
    else:
        start_time_field_name = constants.JOB_START_TIME_DEFAULT_FIELD_NAME

    if constants.JOB_END_TIME_FIELD_NAME in loaded_spec:
        end_time_field_name = loaded_spec[constants.JOB_END_TIME_FIELD_NAME]
    else:
        end_time_field_name = constants.JOB_END_TIME_DEFAULT_FIELD_NAME

    if constants.JOB_PARSER_CLASS_NAME_FIELD in loaded_spec:
        job_parser_class_name = loaded_spec[constants.JOB_PARSER_CLASS_NAME_FIELD]
    else:
        job_parser_class_name = None

    if constants.JOB_ENTRY_FILTER_FIELD in loaded_spec:
        job_entry_filter = loaded_spec[constants.JOB_ENTRY_FILTER_FIELD]
    else:
        job_entry_filter = dict()

    if constants.JOB_INPUTS_FEATURE_ENGINEERING in loaded_spec:
        job_input_feature_engineering = loaded_spec[constants.JOB_INPUTS_FEATURE_ENGINEERING]
    else:
        job_input_feature_engineering = dict()

    if constants.METADATA_PARSER_CLASS_NAME_FIELD in loaded_spec:
        metadata_parser_class_name = loaded_spec[constants.METADATA_PARSER_CLASS_NAME_FIELD]
    else:
        metadata_parser_class_name = None

    return (set(loaded_spec[constants.JOB_INPUTS_FIELD_NAME]), set(loaded_spec[constants.JOB_OUTPUTS_FIELD_NAME]),
            start_time_field_name, end_time_field_name, job_parser_class_name, job_entry_filter,
            job_input_feature_engineering, metadata_parser_class_name)

# This function computes duration fo the job based on start and end time
# It needs to work on each job row separately, because unfortunately different rows report time in different formats


def get_job_duration_from_json(job_data, start_time_field_name, end_time_field_name):

    start_time = utils.find_item(job_data, start_time_field_name)
    end_time = utils.find_item(job_data, end_time_field_name)

    if not start_time or not end_time:
        raise Exception("Both {} and {} should be in historical executions".format(start_time_field_name,
                                                                                   end_time_field_name))

    return utils.get_duration(start_time, end_time)


def process_single_file(filename, columns, class_inst, job_parser_obj):

    if os.path.isfile(filename):
        if filename.endswith(".csv"):
            if class_inst:
                return getattr(class_inst, "get_history_csv")(filename)
            else:
                return pd.read_csv(filename)[columns]
        elif filename.endswith(".json"):
            if class_inst:
                return getattr(class_inst, "get_history_json")(filename)
            else:
                return job_parser_obj.get_history_json(filename)
        else:
            logger.warning("file format of {} not supported, skipping".format(filename))
            return None

    logger.warning("{} is not a file, skipping".format(filename))
    return None


def collect_jobs_history(data_dir, output_path, job_inputs, job_outputs, start_time_field_name,
                         end_time_field_name, input_file=None, job_parser_class_name=None, job_entry_filter={},
                         feature_engineering=None, metadata_parser_class_name=None, metadata_path=None):

    """
    Collects historical executions of jobs into a CSV file according to given job input and output spec. Replaces
    start and end time columns with duration column.

    :param data_dir:
    :param output_path:
    :param job_inputs:
    :param job_outputs:
    :param start_time_field_name:
    :param end_time_field_name:
    :param input_file:
    :param job_parser_class_name:
    :param job_entry_filter
    :param feature_engineering: dict of feature to augment and their corresponding csv file
    :param metadata_parser_class_name class name of the parser for metadata info for feature engineering
    :param metadata_path:path to metadata files
    :return: a dataframe containing the historical data and a file name to which the data was stored. None if parsing
    failed.
    """

    columns_with_derived = utils.adjust_columns_with_duration(job_inputs + job_outputs, start_time_field_name,
                                                              end_time_field_name)
    filter_columns = [] if not job_entry_filter else [entry[constants.JOB_ENTRY_FILTER_NAME_COL] for
                                                      entry in job_entry_filter if not
                                                      entry[constants.JOB_ENTRY_FILTER_KEEP_COL]]
    columns_with_derived = columns_with_derived + filter_columns
    df = pd.DataFrame(columns=columns_with_derived)

    if not os.path.exists(data_dir):
        logger.error("Data directory {} does not exist".format(data_dir))
        return None, None

    if not os.path.exists(output_path):
        logger.error("output path {} does not exist".format(output_path))
        return None, None

    job_parser_obj = None
    class_inst = None

    if job_parser_class_name:
        logger.info("Parsing jobs history with class {}".format(job_parser_class_name))
        estimation_module = __import__('preprocessing.custom_job_parser', fromlist=[job_parser_class_name])
        class_inst = getattr(estimation_module, job_parser_class_name)(columns_with_derived, start_time_field_name,
                                                                       end_time_field_name)
    else:
        job_parser_obj = custom_job_parser.DefaultJsonJobParser(columns_with_derived, start_time_field_name,
                                                                end_time_field_name)

    if input_file is not None:
        df = process_single_file(os.path.join(data_dir, input_file), columns_with_derived, class_inst,
                                 job_parser_obj)
    else:
        for root, subdirs, files in os.walk(data_dir):
            for filename in files:
                # each metadata file may contain more than one job
                new_history_df = process_single_file(os.path.join(root, filename), columns_with_derived, class_inst,
                                                     job_parser_obj)
                if new_history_df is not None:
                    new_history_df.reset_index(drop=True, inplace=True)
                    if df.empty:
                        df = new_history_df
                    else:
                        df.reset_index(drop=True, inplace=True)
                        df = pd.concat([df, new_history_df])
    if df.empty:
        logger.error("No execution history found in {}".format(data_dir))
        return None, None
    else:
        if job_entry_filter:
            for entry in job_entry_filter:
                df = df[~df[entry[constants.JOB_ENTRY_FILTER_NAME_COL]].isin(entry[constants.JOB_ENTRY_FILTER_VALUES_COL])]
                if not entry[constants.JOB_ENTRY_FILTER_KEEP_COL]:
                    df = df.drop(entry[constants.JOB_ENTRY_FILTER_NAME_COL], axis=1)
        logger.info("Found {:d} executions in history".format(len(df)))

    collect_and_persist_data_metadata(df, job_inputs, job_outputs, output_path)

    if feature_engineering:
        df = utils.add_feature_engineering(metadata_path, df, feature_engineering, metadata_parser_class_name)

    output_file = os.path.join(output_path, constants.JOB_HISTORY_FILE_NAME + ".csv")

    df.to_csv(output_file, index=False)
    return df, output_file


def collect_and_persist_data_metadata(df: pd.DataFrame, inputs: List[str], outputs: List[str], output_path: str):

    data_metadata = dict()

    data_metadata['inputs'] = inputs
    data_metadata['outputs'] = outputs

    categorical_features = utils.get_categorical_features(df)

    for job_input in inputs:

        input_metadata = dict()
        input_metadata['type'] = 'categorical' if job_input in categorical_features else 'numeric'
        if input_metadata['type'] == 'categorical':
            input_metadata['values'] = df[job_input].unique().tolist()
        else:
            input_min = df[job_input].min()
            input_max = df[job_input].max()
            input_metadata['min'] = float(input_min) if isinstance(input_min, np.floating) else int(input_min)
            input_metadata['max'] = float(input_max) if isinstance(input_max, np.floating) else int(input_max)

        data_metadata[job_input] = input_metadata

    output_file = os.path.join(output_path, constants.JOB_METADATA_FILE_NAME + ".yaml")

    with open(output_file, 'w') as out:
        yaml.dump(data_metadata, out)



import unittest
import logging
import sys

from tests.utils.logger_redirector import LoggerRedirector
from arise_predictions.preprocessing import job_parser

"""
Tests for job parser.
"""

logger = logging.getLogger('test-logger')

logging.basicConfig(
    format="%(asctime)s %(module)s %(levelname)s: %(message)s",
    level=logging.INFO,
    stream=sys.stdout)


class TestJobParser(unittest.TestCase):

    def setUp(self):
        # unittest has reassigned sys.stdout and sys.stderr by this point
        LoggerRedirector.redirect_loggers(fake_stdout=sys.stdout, 
                                          fake_stderr=sys.stderr)

    def test_get_job_outputs(self):
        """
        Test that we can get the job outputs parsed from job_spec.yaml
        """
        loaded_job_spec = job_parser.parse_job_spec(
            "tests/resources/job_spec.yaml")
        target_variables = sorted(list(loaded_job_spec[1]))
        self.assertTrue(1, len(target_variables))
        self.assertIn("tokens_per_second", target_variables)



import unittest
import logging
import sys
import os
import joblib

import pandas as pd

from arise_predictions.utils import utils, constants
from tests.utils.logger_redirector import LoggerRedirector
from arise_predictions.auto_model.build_models import (get_estimators_config,
                                           _init_estimators,
                                           _search_models,
                                           _rank_estimators,
                                           _select_best_estimators)

from arise_predictions.cmd.cmd import parse_args

"""
Tests for auto model building functionality.

We use an auto-model configuration with two models for two target variables for
this test. Each model comes out as the winner. However, I wonder whether it is
possible to get different results once every so many executions that flip the
results or lead to one model being the winner for both target variables. Have
not encountered this, we are setting the seed for models that allow this and
when splitting the data. So probably unlikely, but something to keep in mind.

Also note that the tearDown method deletes any temporary test files that may be
useful for debugging. Have left shutil.rmtree commented out to avoid this.
"""

logger = logging.getLogger('test-logger')

logging.basicConfig(
    format="%(asctime)s %(module)s %(levelname)s: %(message)s",
    level=logging.INFO,
    stream=sys.stdout)


class TestBuildModels(unittest.TestCase):

    def setUp(self):
        # unittest has reassigned sys.stdout and sys.stderr by this point
        LoggerRedirector.redirect_loggers(fake_stdout=sys.stdout, 
                                          fake_stderr=sys.stderr) 
        
        # Simulate command-line arguments
        parse_args(["--num-jobs", "-1"])
        
        self.resources_path = "tests/resources"
        self.config_file = os.path.join(self.resources_path, "model-search-config.yaml")
        self.data_file = "mlcommons_inference_data_execution_history.csv"
        self.target_variables = ["tokens_per_second"]

        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        self.categorical_variables: list[str] = utils.get_categorical_features(data)
        self.categorical_variables_indices = [data.columns.get_loc(var) for var in self.categorical_variables]

        self.script_dir = os.path.dirname(__file__)
        self.output_path = os.path.abspath(os.path.join(
               self.script_dir, "test-tmp/ARISE-auto-models"))

        # self.best_model_gpu = "model-Ridge-gpu_memory_utilization_max.pkl"
        # self.best_model_time = "model-LinearRegression-train_runtime.pkl"
        # self.best_params_gpu = "params-Ridge-gpu_memory_utilization_max.yaml"
        # self.best_params_time = "params-LinearRegression-train_runtime.yaml"
        self.train_data = "train-data-tokens_per_second.csv"

        self.results_ridge = ("cv_results-RidgeRegression-tokens_per_second"
                                  ".csv")
        self.results_lr = "cv_results-LinearRegression-tokens_per_second.csv"
        self.results_xgb = "cv_results-XGBoost-Regressor-tokens_per_second.csv"
        
        self.rankings = "all-star-rankings.csv"

        self.estimator_linear = "estimator-linear-RidgeRegression-tokens_per_second.pkl"
        self.estimator_nonlinear = "estimator-nonlinear-XGBoost-Regressor-tokens_per_second.pkl"

        self.predictions_lr = "predictions-LinearRegression-tokens_per_second.csv"
        self.predictions_ridge = "predictions-RidgeRegression-tokens_per_second.csv"
        self.predictions_xgb = "predictions-XGBoost-Regressor-tokens_per_second.csv"

        self.best_estimators_summary = "best-estimators-testset-performance-summary.csv"

        self.feature_col = "# of Accelerators"
        self.high_threshold = 6
        self.extrapolation_file = "extrapolation-test-data.csv"

    def tearDown(self):
        LoggerRedirector.reset_loggers(
            fake_stdout=sys.stdout, fake_stderr=sys.stderr)
        
        # if os.path.exists(self.output_path):
        #     shutil.rmtree(self.output_path)

    def test_reads_estimators_and_params_from_config(self):
        """
        Reads estimators and their hyperparameters from the configuration file.
        """
        config = get_estimators_config(self.config_file)
        estimators = config.estimators
        logger.error(f"estimators from YAML:\n{estimators}")
        self.assertEqual(4, len(estimators))
        self.assertTrue(any(d.name == "RidgeRegression" for d in estimators))
    
    def test_setup_estimators(self):
        """
        List of estimators is instantiated according to configuration.
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
            config, self.categorical_variables_indices)
        self.assertEqual(len(estimators), 
                         len(config.estimators))
        estimator_names = [x[0] for x in estimators]
        config_names = [x.name for x in config.estimators]
        logger.error(f"estimator names: {estimator_names}")
        logger.error(f"config names: {config_names}")
        self.assertTrue(set(estimator_names) == set(config_names))
    
    def test_training_files_exist(self):
        """
        CSV files containing training data have been written.
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
            config, self.categorical_variables_indices)
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        _search_models(data=data, estimators=estimators, config=config,
                       categorical_variables=self.categorical_variables,
                       target_variables=self.target_variables, 
                       output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.train_data
        )))

    def test_search_output_exists(self):
        """
        CSV files for detailed search CV results are written as expected.
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
            config, self.categorical_variables_indices)
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        _search_models(data=data, estimators=estimators, config=config,
                       categorical_variables=self.categorical_variables,
                       target_variables=self.target_variables, 
                       output_path=self.output_path)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, 
            self.results_ridge)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, 
            self.results_lr)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path, 
            self.results_xgb)))

    def test_rankings_output_exists(self):
        """
        Rankings CSV file with model performance per target variable exists. 
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
            config, self.categorical_variables_indices)
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        stats_df, _ = _search_models(
            data=data, estimators=estimators, 
            config=config,
            categorical_variables=self.categorical_variables,
            target_variables=self.target_variables, 
            output_path=self.output_path)
        _rank_estimators(
            summary_stats=stats_df, 
            output_path=self.output_path,
            output_file=constants.AM_RANKINGS_FILE)
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.rankings)))
        
    def test_estimators_and_stuff_exists(self):
        """
        Estimators and test set results have been persisted
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
            config, cat_indices=self.categorical_variables_indices)
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        stats_df, estimators_per_target_variable = _search_models(
            data=data, estimators=estimators, 
            config=config,
            categorical_variables=self.categorical_variables,
            target_variables=self.target_variables, 
            output_path=self.output_path)
        rankings = _rank_estimators(
            summary_stats=stats_df, 
            output_path=self.output_path,
            output_file=constants.AM_RANKINGS_FILE)
        _select_best_estimators(
            target_variables=self.target_variables,
            rankings=rankings,
            estimators_per_target_variable=estimators_per_target_variable,
            output_path=self.output_path,
            num_jobs=config.num_jobs,
            leave_one_out_cv=None,
            categorical_variables=self.categorical_variables)

        # best linear and non-linear models for gpu target variable
        # print(f"******* self.output_path: {self.output_path}")
        self.assertTrue(os.path.exists(self.output_path))

        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.estimator_linear)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.estimator_nonlinear)))

        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.predictions_ridge)))
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.predictions_xgb)))
        
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.best_estimators_summary)))
        
    def test_load_and_predict(self):
        """
        Test that we can load a persisted estimator and use it to predict.
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
            config, self.categorical_variables_indices)
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        stats_df, estimators_per_target_variable = _search_models(
            data=data, estimators=estimators, 
            config=config,
            categorical_variables=self.categorical_variables,
            target_variables=self.target_variables, 
            output_path=self.output_path)
        rankings = _rank_estimators(
            summary_stats=stats_df, 
            output_path=self.output_path,
            output_file=constants.AM_RANKINGS_FILE)
        _select_best_estimators(
            target_variables=self.target_variables,
            rankings=rankings,
            estimators_per_target_variable=estimators_per_target_variable,
            output_path=self.output_path,
            num_jobs=config.num_jobs,
            leave_one_out_cv=None,
            categorical_variables=self.categorical_variables)

        test_estimator_path = os.path.join(
            self.output_path, 
            self.estimator_linear)
        loaded_estimator = joblib.load(test_estimator_path)
        X = data.drop(self.target_variables, axis=1)
        y_pred = loaded_estimator.predict(X)
        self.assertTrue(len(y_pred) > 0)
    
    def test_extrapolation_data_exists_if_feature_col(self):
        """
        Extrapolation test set created if feature column has been set.
        """
        config = get_estimators_config(self.config_file)
        estimators = _init_estimators(
                config, cat_indices=self.categorical_variables_indices)
        data = pd.read_csv(os.path.join(self.resources_path, self.data_file))
        stats_df, estimators_per_target_variable = _search_models(
            data=data, estimators=estimators, 
            config=config,
            categorical_variables=self.categorical_variables,
            target_variables=self.target_variables, 
            output_path=self.output_path,
            feature_col=self.feature_col,
            high_threshold=self.high_threshold)
        
        self.assertTrue(os.path.exists(os.path.join(
            self.output_path,
            self.extrapolation_file)))

        df = pd.read_csv(os.path.join(
            self.output_path,
            self.extrapolation_file
        ))
        self.assertGreater(len(df), 0, 
                           "Extrapolation DF should contain at least 1 row.")
        self.assertTrue(all(df[self.feature_col] >= self.high_threshold),
                            (f"All values in {self.feature_col} should be at" 
                             f" least {self.high_threshold}"))


from setuptools import setup, find_packages
from glob import glob
import arise_predictions

setup(
    name='arise_predictions',
    version=arise_predictions.__version__,
    packages=find_packages(),
    license='Apache License 2.0',
    description='ML-based performance prediction for AI workloads configuration management',
    long_description='ML-based performance prediction for AI workloads configuration management',
    python_requires='>=3.9',
    install_requires=[
        'pandas~=2.1.3',
        'PyYaml',
        'scikit-learn==1.5.1',
        'numpy',
        'scipy',
        'numpyencoder',
        'linear-tree',
        'stopit',
        'kneed',
        'ConfigArgParse',
        'matplotlib',
        'xgboost==2.1.1',
        'catboost',
    ],
    data_files=[('tests', glob('tests/**/*.*', recursive=True))],
)
